# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Sauter conductivity model."""
import dataclasses
from typing import Annotated, Literal

import jax
import jax.numpy as jnp
from torax._src import array_typing
from torax._src import jax_utils
from torax._src import state
from torax._src.fvm import cell_variable
from torax._src.geometry import geometry as geometry_lib
from torax._src.neoclassical import formulas
from torax._src.neoclassical.conductivity import base
from torax._src.neoclassical.conductivity import runtime_params
from torax._src.physics import collisions
from torax._src.torax_pydantic import torax_pydantic


# TODO(b/425750357): Add neoclassical correciton flag (default to True)
@jax.tree_util.register_dataclass
@dataclasses.dataclass(frozen=True)
class RuntimeParams(runtime_params.RuntimeParams):
  """Runtime params for the Sauter model."""


class SauterModel(base.ConductivityModel):
  """Sauter conductivity model."""

  def calculate_conductivity(
      self,
      geometry: geometry_lib.Geometry,
      core_profiles: state.CoreProfiles,
  ) -> base.Conductivity:
    """Calculates conductivity."""
    result = _calculate_conductivity(
        Z_eff_face=core_profiles.Z_eff_face,
        n_e=core_profiles.n_e,
        T_e=core_profiles.T_e,
        q_face=core_profiles.q_face,
        geo=geometry,
    )
    return base.Conductivity(
        sigma=result.sigma,
        sigma_face=result.sigma_face,
    )

  def __eq__(self, other) -> bool:
    return isinstance(other, self.__class__)

  def __hash__(self) -> int:
    return hash(self.__class__)


class SauterModelConfig(base.ConductivityModelConfig):
  """Sauter conductivity model config."""

  model_name: Annotated[Literal['sauter'], torax_pydantic.JAX_STATIC] = 'sauter'

  def build_runtime_params(self) -> RuntimeParams:
    return RuntimeParams()

  def build_model(self) -> SauterModel:
    return SauterModel()


@jax_utils.jit
def _calculate_conductivity(
    *,
    Z_eff_face: array_typing.FloatVectorFace,
    n_e: cell_variable.CellVariable,
    T_e: cell_variable.CellVariable,
    q_face: array_typing.FloatVectorFace,
    geo: geometry_lib.Geometry,
) -> base.Conductivity:
  """Calculates sigma and sigma_face using the Sauter model."""
  # pylint: disable=invalid-name

  # Formulas from Sauter PoP 1999.

  # Effective trapped particle fraction
  f_trap = formulas.calculate_f_trap(geo)

  # Spitzer conductivity
  NZ = 0.58 + 0.74 / (0.76 + Z_eff_face)
  log_lambda_ei = collisions.calculate_log_lambda_ei(
      T_e.face_value(), n_e.face_value()
  )

  sigsptz = (
      1.9012e04
      * (T_e.face_value() * 1e3) ** 1.5
      / Z_eff_face
      / NZ
      / log_lambda_ei
  )

  nu_e_star_face = formulas.calculate_nu_e_star(
      q=q_face,
      geo=geo,
      n_e=n_e.face_value(),
      T_e=T_e.face_value(),
      Z_eff=Z_eff_face,
      log_lambda_ei=log_lambda_ei,
  )

  # Neoclassical correction to spitzer conductivity
  ft33 = f_trap / (
      1.0
      + (0.55 - 0.1 * f_trap) * jnp.sqrt(nu_e_star_face)
      + 0.45 * (1.0 - f_trap) * nu_e_star_face / (Z_eff_face**1.5)
  )
  signeo_face = 1.0 - ft33 * (
      1.0
      + 0.36 / Z_eff_face
      - ft33 * (0.59 / Z_eff_face - 0.23 / Z_eff_face * ft33)
  )
  sigma_face = sigsptz * signeo_face

  sigmaneo_cell = geometry_lib.face_to_cell(sigma_face)

  return base.Conductivity(
      sigma=sigmaneo_cell,
      sigma_face=sigma_face,
  )



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Plotting configuration for Torax runs focusing on transport coefficients."""

from torax._src.plotting import plotruns_lib

PLOT_CONFIG = plotruns_lib.FigureProperties(
    rows=2,
    cols=3,
    tick_fontsize=10,
    axes_fontsize=10,
    default_legend_fontsize=9,
    figure_size_factor=8,
    title_fontsize=12,
    axes=(
        plotruns_lib.PlotProperties(
            attrs=(
                'chi_turb_i',
                'chi_turb_e',
            ),
            labels=(
                r'$\chi_\mathrm{turb,i}$',
                r'$\chi_\mathrm{turb,e}$',
            ),
            ylabel=r'Turbulent heat conductivity $[m^2/s]$',
            upper_percentile=98.0,
            include_first_timepoint=False,
            ylim_min_zero=False,
        ),
        plotruns_lib.PlotProperties(
            attrs=('D_turb_e',),
            labels=(r'$D_\mathrm{turb,e}$',),
            ylabel=r'Turbulent particle diffusivity $[m^2/s]$',
            upper_percentile=98.0,
            lower_percentile=2.0,
            include_first_timepoint=False,
            ylim_min_zero=False,
        ),
        plotruns_lib.PlotProperties(
            attrs=('V_turb_e',),
            labels=(r'$V_\mathrm{turb,e}$',),
            ylabel=r'Turbulent particle convectivity $[m^2/s]$',
            upper_percentile=98.0,
            lower_percentile=2.0,
            include_first_timepoint=False,
            ylim_min_zero=False,
        ),
        plotruns_lib.PlotProperties(
            attrs=(
                'chi_neo_i',
                'chi_neo_e',
            ),
            labels=(
                r'$\chi_\mathrm{neo,i}$',
                r'$\chi_\mathrm{neo,e}$',
            ),
            ylabel=r'Neoclassical heat conductivity $[m^2/s]$',
            upper_percentile=98.0,
            include_first_timepoint=False,
            ylim_min_zero=False,
            suppress_zero_values=True,
        ),
        plotruns_lib.PlotProperties(
            attrs=('D_neo_e',),
            labels=(r'$D_\mathrm{neo,e}$',),
            ylabel=r'Neoclassical particle diffusivity $[m^2/s]$',
            upper_percentile=98.0,
            lower_percentile=2.0,
            include_first_timepoint=False,
            ylim_min_zero=False,
            suppress_zero_values=True,
        ),
        plotruns_lib.PlotProperties(
            attrs=('V_neo_total_e', 'V_neo_e', 'V_neo_ware_e'),
            labels=(
                r'$V_\mathrm{neo,e}+V_\mathrm{ware,e}$',
                r'$V_\mathrm{neo,e}$',
                r'$V_\mathrm{ware,e}$',
            ),
            ylabel=r'Neoclassical particle convectivity $[m^2/s]$',
            upper_percentile=98.0,
            lower_percentile=2.0,
            include_first_timepoint=False,
            ylim_min_zero=False,
            suppress_zero_values=True,
        ),
    ),
)



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Classes and functions for defining interpolated parameters."""

from collections.abc import Mapping
import functools
from typing import Any, Literal, TypeAlias
import chex
import numpy as np
import pydantic
from torax._src import array_typing
from torax._src import interpolated_param
from torax._src import jax_utils
from torax._src.torax_pydantic import model_base
from torax._src.torax_pydantic import pydantic_types
import typing_extensions
import xarray as xr

ValueType: TypeAlias = dict[
    float,
    tuple[pydantic_types.NumpyArray1DUnitInterval, pydantic_types.NumpyArray1D],
]


class Grid1D(model_base.BaseModelFrozen):
  """Data structure defining a 1-D grid of cells with faces.

  Attributes:
    nx: Number of cells.
  """

  nx: typing_extensions.Annotated[pydantic.conint(ge=4), model_base.JAX_STATIC]

  @functools.cached_property
  def dx(self) -> float:
    return 1 / self.nx

  @property
  def face_centers(self) -> np.ndarray:
    """Coordinates of face centers."""
    return _get_face_centers(nx=self.nx, dx=self.dx)

  @property
  def cell_centers(self) -> np.ndarray:
    """Coordinates of cell centers."""
    return _get_cell_centers(nx=self.nx, dx=self.dx)

  def __eq__(self, other: typing_extensions.Self) -> bool:
    return self.nx == other.nx and self.dx == other.dx

  def __hash__(self) -> int:
    return hash((self.nx, self.dx))


class TimeVaryingArray(model_base.BaseModelFrozen):
  """Base class for time interpolated array types.

  The Pydantic `.model_validate` constructor can accept a variety of input types
  defined by the `TimeRhoInterpolatedInput` type. See
  https://torax.readthedocs.io/en/latest/configuration.html#time-varying-arrays
  for more details.

  Attributes:
    value: A dict of the form `{time: (rho_norm, values), ...}`, where
      `rho_norm` and `values` are 1D NumPy arrays, and each pair is of equal
      length. Note that all `rho_norm` values `x` are in the range `0 <= x <=
      1.0`, and the mapping is ordered by increasing `time`.
    rho_interpolation_mode: The interpolation mode to use for the rho axis.
    time_interpolation_mode: The interpolation mode to use for the time axis.
    grid_face_centers: The face centers of the grid to use for the
      interpolation. This is optional, as this value is often not known at
      construction time, and is set later.
    grid_cell_centers: The cell centers of the grid to use for the
      interpolation. This is optional, as this value is often not known at
      construction time, and is set later.
  """

  value: ValueType
  rho_interpolation_mode: typing_extensions.Annotated[
      interpolated_param.InterpolationMode, model_base.JAX_STATIC
  ] = interpolated_param.InterpolationMode.PIECEWISE_LINEAR
  time_interpolation_mode: typing_extensions.Annotated[
      interpolated_param.InterpolationMode, model_base.JAX_STATIC
  ] = interpolated_param.InterpolationMode.PIECEWISE_LINEAR
  grid: Grid1D | None = None

  def tree_flatten(self):
    children = (
        self.value,
        # Save out the cached interpolated params.
        self._get_cached_interpolated_param_cell,
        self._get_cached_interpolated_param_face,
        self._get_cached_interpolated_param_face_right,
    )
    aux_data = (
        self.rho_interpolation_mode,
        self.time_interpolation_mode,
        self.grid,
    )
    return children, aux_data

  @classmethod
  def tree_unflatten(cls, aux_data, children):
    # Avoid calling model_validate as validation should be done already.
    obj = cls.model_construct(
        value=children[0],
        rho_interpolation_mode=aux_data[0],
        time_interpolation_mode=aux_data[1],
        grid=aux_data[2],
    )
    # Plug back in the cached interpolated params to avoid losing the cache.
    # pylint: disable=protected-access
    obj._get_cached_interpolated_param_cell = children[1]
    obj._get_cached_interpolated_param_face = children[2]
    obj._get_cached_interpolated_param_face_right = children[3]
    # pylint: enable=protected-access
    return obj

  @functools.cached_property
  def right_boundary_conditions_defined(self) -> bool:
    """Checks if the boundary condition at rho=1.0 is always defined."""

    for rho_norm, _ in self.value.values():
      if 1.0 not in rho_norm:
        return False
    return True

  def get_value(
      self,
      t: chex.Numeric,
      grid_type: Literal['cell', 'face', 'face_right'] = 'cell',
  ) -> array_typing.Array:
    """Returns the value of this parameter interpolated at x=time.

    Args:
      t: An array of times to interpolate at.
      grid_type: One of 'cell', 'face', or 'face_right'. For 'face_right', the
        element `self.grid_face_centers[-1]` is used as the grid.

    Raises:
      RuntimeError: If `self.grid` is None.

    Returns:
      An array of interpolated values.
    """
    match grid_type:
      case 'cell':
        return self._get_cached_interpolated_param_cell.get_value(t)
      case 'face':
        return self._get_cached_interpolated_param_face.get_value(t)
      case 'face_right':
        return self._get_cached_interpolated_param_face_right.get_value(t)
      case _:
        raise ValueError(f'Unknown grid type: {grid_type}')

  def __eq__(self, other: typing_extensions.Self):
    try:
      chex.assert_trees_all_equal(self.value, other.value)
      return (
          self.rho_interpolation_mode == other.rho_interpolation_mode
          and self.time_interpolation_mode == other.time_interpolation_mode
          and self.grid == other.grid
      )
    except AssertionError:
      return False

  @pydantic.field_validator('value', mode='after')
  @classmethod
  def _valid_value(cls, value: ValueType) -> ValueType:
    # Ensure the keys are sorted.
    value = dict(sorted(value.items()))

    for t, (rho_norm, values) in value.items():
      if not isinstance(t, float):
        raise ValueError(f'Time values must be a float, but got {t}.')

      if not np.issubdtype(rho_norm.dtype, np.floating):
        raise ValueError(
            f'rho_norm must be a float array, but got {rho_norm.dtype}.'
        )
      if rho_norm.dtype != values.dtype:
        raise ValueError(
            'rho_norm and values must have the same dtype. Given: '
            f'{rho_norm.dtype} and {values.dtype}.'
        )
      if len(rho_norm) != len(values):
        raise ValueError(
            'rho_norm and values must be of the same length. Given: '
            f'{len(rho_norm)} and {len(values)}.'
        )

    return value

  @pydantic.model_validator(mode='before')
  @classmethod
  def _conform_data(
      cls, data: interpolated_param.TimeRhoInterpolatedInput | dict[str, Any]
  ) -> dict[str, Any]:

    if isinstance(data, dict):
      # A workaround for https://github.com/pydantic/pydantic/issues/10477.
      data.pop('_get_cached_interpolated_param_cell_centers', None)
      data.pop('_get_cached_interpolated_param_face_centers', None)
      data.pop('_get_cached_interpolated_param_face_right_centers', None)

      # This is the standard constructor input. No conforming required.
      if set(data.keys()).issubset(cls.model_fields.keys()):
        return data

    # Potentially parse the interpolation modes from the input.
    time_interpolation_mode = (
        interpolated_param.InterpolationMode.PIECEWISE_LINEAR
    )
    rho_interpolation_mode = (
        interpolated_param.InterpolationMode.PIECEWISE_LINEAR
    )

    if isinstance(data, tuple):
      if len(data) == 2 and isinstance(data[1], dict):
        time_interpolation_mode = interpolated_param.InterpolationMode[
            data[1]['time_interpolation_mode'].upper()
        ]
        rho_interpolation_mode = interpolated_param.InterpolationMode[
            data[1]['rho_interpolation_mode'].upper()
        ]
        # First element in tuple assumed to be the input.
        data = data[0]

    if isinstance(data, xr.DataArray):
      value = _load_from_arrays(data)
    elif isinstance(data, tuple):
      values = []
      for v in data:
        if isinstance(v, array_typing.Array):
          values.append(v)
        elif isinstance(v, list):
          values.append(np.asarray(v))
        else:
          raise ValueError(
              'Input to TimeVaryingArray unsupported. Input was of type:'
              f' {type(v)}. Expected array_typing.Array or list of'
              ' floats/ints/bools.'
          )
      value = _load_from_arrays(tuple(values))
    elif isinstance(data, Mapping) or isinstance(data, (float, int)):
      value = _load_from_primitives(data)
    else:
      raise ValueError(
          'Input to TimeVaryingArray unsupported. Input was of type:'
          f' {type(data)}'
      )

    return dict(
        value=value,
        time_interpolation_mode=time_interpolation_mode,
        rho_interpolation_mode=rho_interpolation_mode,
    )

  @functools.cached_property
  def _get_cached_interpolated_param_cell(
      self,
  ) -> interpolated_param.InterpolatedVarTimeRho:
    if self.grid is None:
      raise RuntimeError('grid must be set.')

    return interpolated_param.InterpolatedVarTimeRho(
        self.value,
        rho_norm=self.grid.cell_centers,
        time_interpolation_mode=self.time_interpolation_mode,
        rho_interpolation_mode=self.rho_interpolation_mode,
    )

  @functools.cached_property
  def _get_cached_interpolated_param_face(
      self,
  ) -> interpolated_param.InterpolatedVarTimeRho:
    if self.grid is None:
      raise RuntimeError('grid must be set.')

    return interpolated_param.InterpolatedVarTimeRho(
        self.value,
        rho_norm=self.grid.face_centers,
        time_interpolation_mode=self.time_interpolation_mode,
        rho_interpolation_mode=self.rho_interpolation_mode,
    )

  @functools.cached_property
  def _get_cached_interpolated_param_face_right(
      self,
  ) -> interpolated_param.InterpolatedVarTimeRho:
    if self.grid is None:
      raise RuntimeError('grid must be set.')

    return interpolated_param.InterpolatedVarTimeRho(
        self.value,
        rho_norm=self.grid.face_centers[-1],
        time_interpolation_mode=self.time_interpolation_mode,
        rho_interpolation_mode=self.rho_interpolation_mode,
    )


def _is_positive(array: TimeVaryingArray) -> TimeVaryingArray:
  for _, value in array.value.values():
    if not np.all(value > 0):
      raise ValueError('All values must be positive.')
  return array


PositiveTimeVaryingArray = typing_extensions.Annotated[
    TimeVaryingArray, pydantic.AfterValidator(_is_positive)
]


def _load_from_primitives(
    primitive_values: (
        Mapping[float, interpolated_param.InterpolatedVarSingleAxisInput]
        | float
    ),
) -> Mapping[float, tuple[array_typing.Array, array_typing.Array]]:
  """Loads the data from primitives.

  Three cases are supported:
  1. A float is passed in, describes constant initial condition profile.
  2. A non-nested dict is passed in, it will describe the radial profile for
     the initial condition.
  3. A nested dict is passed in, it will describe a time-dependent radial
  profile providing both initial condition and prescribed values at times beyond

  Args:
    primitive_values: The python primitive values to load.

  Returns:
    A mapping from time to (rho_norm, values) where rho_norm and values are both
    arrays of equal length.
  """
  # Float case.
  if isinstance(primitive_values, (float, int)):
    primitive_values = {0.0: {0.0: primitive_values}}
  # Non-nested dict.
  if isinstance(primitive_values, Mapping) and all(
      isinstance(v, float) for v in primitive_values.values()
  ):
    primitive_values = {0.0: primitive_values}

  if len(set(primitive_values.keys())) != len(primitive_values):
    raise ValueError('Indicies in values mapping must be unique.')
  if not primitive_values:
    raise ValueError('Values mapping must not be empty.')

  loaded_values = {}
  for t, v in primitive_values.items():
    x, y, _, _ = interpolated_param.convert_input_to_xs_ys(v)
    loaded_values[t] = (x, y)

  return loaded_values


def _load_from_arrays(
    arrays: tuple[array_typing.Array, ...] | xr.DataArray,
) -> Mapping[float, tuple[np.ndarray, np.ndarray]]:
  """Loads the data from numpy arrays.

  Args:
    arrays: A tuple of (times, rho_norm, values) or (rho_norm, values), or an
      xarray.DataArray.

  Returns:
    A mapping from time to (rho_norm, values)
  """

  if isinstance(arrays, xr.DataArray):
    if interpolated_param.RHO_NORM not in arrays.coords:
      raise ValueError(
          f'"{interpolated_param.RHO_NORM}" must be a coordinate in given'
          ' dataset.'
      )
    if 'time' in arrays.coords:
      arrays = (arrays.time.data, arrays.rho_norm.data, arrays.data)
    else:
      arrays = (arrays.rho_norm.data, arrays.data)

  if len(arrays) == 2:
    # Shortcut for initial condition profile.
    rho_norm, values = arrays  # pytype: disable=bad-unpacking
    return {
        0.0: (
            np.asarray(rho_norm, dtype=jax_utils.get_np_dtype()),
            np.asarray(values, dtype=jax_utils.get_np_dtype()),
        )
    }
  if len(arrays) == 3:
    times = np.asarray(arrays[0], dtype=jax_utils.get_np_dtype())
    rho_norm = np.asarray(arrays[1], dtype=jax_utils.get_np_dtype())
    values = np.asarray(arrays[2], dtype=jax_utils.get_np_dtype())

    if values.ndim != 2:
      raise ValueError(
          f'The values array must have ndim=2, but got {values.ndim}.'
      )

    if rho_norm.ndim == 1:
      rho_norm = np.stack([rho_norm] * len(times))

    return {t: (rho_norm[i], values[i]) for i, t in enumerate(times)}
  else:
    raise ValueError(f'arrays must be length 2 or 3. Given: {len(arrays)}.')


def set_grid(
    model: model_base.BaseModelFrozen,
    grid: Grid1D,
    mode: Literal['strict', 'force', 'relaxed'] = 'strict',
):
  """Sets the grid for for the model and all its submodels.

  Args:
    model: The model to set the geometry mesh for.
    grid: A `Grid1D` object.
    mode: The update mode. With `'strict'` (default), an error will be raised if
      the `grid` is already set in `model` or any of its submodels. With
      `'force'`, `grid` will be overwritten in a potentially unsafe way (no
      cache invalidation). With `'relaxed'`, `grid` will only be set if it is
      not already set.

  Raises:
    RuntimeError: If `force_update=False` and `grid` is already set.
  """

  def _update_rule(submodel):
    # The update API assumes all submodels are unique objects. Construct
    # a new Grid1D object (without validation) to ensure this. We do reuse
    # the same NumPy arrays.
    new_grid = Grid1D.model_construct(
        nx=grid.nx,
        face_centers=grid.face_centers,
        cell_centers=grid.cell_centers,
    )
    if submodel.grid is None:
      submodel.__dict__['grid'] = new_grid
    else:
      match mode:
        case 'force':
          submodel.__dict__['grid'] = new_grid
        case 'relaxed':
          pass
        case 'strict':
          raise RuntimeError(
              '`grid` is already set and using strict update mode.'
          )

  for submodel in model.submodels:
    if isinstance(submodel, TimeVaryingArray):
      _update_rule(submodel)


def _is_non_negative(
    time_varying_array: TimeVaryingArray,
) -> TimeVaryingArray:
  for _, value in time_varying_array.value.values():
    if not np.all(value >= 0.0):
      raise ValueError('All values must be non-negative.')
  return time_varying_array


# The Torax mesh objects will generally have the same grid parameters. Thus
# a global cache prevents recomputing the same linspaces for each mesh.
@functools.cache
def _get_face_centers(nx: int, dx: float) -> np.ndarray:
  return np.linspace(0, nx * dx, nx + 1)


@functools.cache
def _get_cell_centers(nx: int, dx: float) -> np.ndarray:
  return np.linspace(dx * 0.5, (nx - 0.5) * dx, nx)

NonNegativeTimeVaryingArray: TypeAlias = typing_extensions.Annotated[
    TimeVaryingArray, pydantic.AfterValidator(_is_non_negative)
]



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Ion mixture model and impurity fractions model for plasma composition."""

from collections.abc import Mapping
import dataclasses
from typing import Annotated, Any, Literal, TypeAlias

import chex
import jax
from jax import numpy as jnp
import jaxtyping as jt
import numpy as np
import pydantic
from torax._src import array_typing
from torax._src import constants
from torax._src.torax_pydantic import torax_pydantic
from typing_extensions import Final


# pylint: disable=invalid-name
_IMPURITY_MODE_FRACTIONS: Final[str] = 'fractions'


def _impurity_before_validator(value: Any) -> Any:
  """Validates the input for the ImpurityMapping."""
  if isinstance(value, str):
    return {value: 1.0}
  return value


def _impurity_after_validator(
    value: Mapping[str, torax_pydantic.TimeVaryingArray],
) -> Mapping[str, torax_pydantic.TimeVaryingArray]:
  """Validates a dictionary of TimeVaryingArray objects that form a composition.

  This validator checks for three conditions:
  1. All TimeVaryingArray objects have the same time points.
  2. For each time point, all TimeVaryingArray objects have the same `rho_norm`
     array.
  3. For each time point, the `values` arrays across all TimeVaryingArray
     objects have the same shape and sum to 1.0 along the axis of the dictionary
     keys. This is useful for ensuring that fractions of a quantity sum to 1.

  Args:
    value: The dictionary of TimeVaryingArray objects to validate.

  Returns:
    The validated dictionary.

  Raises:
    ValueError: If any of the validation checks fail.
  """
  if not value:
    raise ValueError('The species dictionary cannot be empty.')

  first_key = next(iter(value))
  first_tva = value[first_key]
  reference_times = first_tva.value.keys()

  for species_key, tva in value.items():
    if tva.value.keys() != reference_times:
      raise ValueError(
          f'Inconsistent times for key "{species_key}". Expected'
          f' {sorted(list(reference_times))}, got'
          f' {sorted(list(tva.value.keys()))}'
      )

  for t in reference_times:
    reference_rho_norm, _ = first_tva.value[t]
    for species_key, tva in value.items():
      current_rho_norm, _ = tva.value[t]
      if not np.array_equal(current_rho_norm, reference_rho_norm):
        raise ValueError(
            f'Inconsistent rho_norm for key "{species_key}" at time {t}.'
        )

    # Check for each time point, `values` have the same shape and sum to 1.0.
    values_at_t = [tva.value[t][1] for tva in value.values()]
    reference_shape = values_at_t[0].shape
    for i, v_arr in enumerate(values_at_t):
      if v_arr.shape != reference_shape:
        species_key = list(value.keys())[i]
        raise ValueError(
            f'Inconsistent value shape for key "{species_key}" at time {t}.'
            f' Expected {reference_shape}, got {v_arr.shape}.'
        )
    sum_of_values = np.sum(np.stack(values_at_t, axis=0), axis=0)
    if not np.allclose(sum_of_values, 1.0):
      raise ValueError(
          f'Values do not sum to 1 at time {t}. Sum is {sum_of_values}.'
      )
  return value


ImpurityMapping: TypeAlias = Annotated[
    Mapping[str, torax_pydantic.NonNegativeTimeVaryingArray],
    pydantic.BeforeValidator(_impurity_before_validator),
    pydantic.AfterValidator(_impurity_after_validator),
]


@jax.tree_util.register_dataclass
@dataclasses.dataclass(frozen=True)
class RuntimeParams:
  """Represents a fixed mixture of ion species at a specific time.

  Information on ion names are not stored here, but rather as static attributes,
  to simplify JAX logic and performance in source functions for fusion power and
  radiation which are species-dependent.

  Attributes:
    fractions: Impurity fractions for a time slice.
    fractions_face: Impurity fractions for a time slice for the face grid.
    A_avg: Average atomic mass of the mixture.
    A_avg_face: Average atomic mass of the mixture for the face grid.
    Z_override: Typically, the average Z is calculated according to the
      temperature dependent charge-state-distribution, or for low-Z cases by the
      atomic numbers of the ions assuming full ionization. If Z_override is
      provided, it is used instead for the average Z.
  """

  fractions: jt.Float[array_typing.Array, 'ion_symbol rhon']
  fractions_face: jt.Float[array_typing.Array, 'ion_symbol rhon+1']
  A_avg: array_typing.FloatVectorCell
  A_avg_face: array_typing.FloatVectorFace
  Z_override: array_typing.FloatScalar | None = None


class ImpurityFractions(torax_pydantic.BaseModelFrozen):
  """Impurity content defined by fractional abundances."""

  impurity_mode: Annotated[Literal['fractions'], torax_pydantic.JAX_STATIC] = (
      'fractions'
  )
  species: ImpurityMapping = torax_pydantic.ValidatedDefault({'Ne': 1.0})
  Z_override: torax_pydantic.TimeVaryingScalar | None = None
  A_override: torax_pydantic.TimeVaryingScalar | None = None

  def build_runtime_params(self, t: chex.Numeric) -> RuntimeParams:
    """Builds a RuntimeParams object at a given time."""
    ions = self.species.keys()
    fractions = jnp.array([self.species[ion].get_value(t) for ion in ions])
    fractions_face = jnp.array(
        [self.species[ion].get_value(t, grid_type='face') for ion in ions]
    )
    Z_override = None if not self.Z_override else self.Z_override.get_value(t)

    if not self.A_override:
      As = jnp.array([constants.ION_PROPERTIES_DICT[ion].A for ion in ions])
      A_avg = jnp.sum(As[..., jnp.newaxis] * fractions, axis=0)
      A_avg_face = jnp.sum(As[..., jnp.newaxis] * fractions_face, axis=0)
    else:
      A_avg = jnp.full_like(fractions[0], self.A_override.get_value(t))
      A_avg_face = jnp.full_like(
          fractions_face[0], self.A_override.get_value(t)
      )

    return RuntimeParams(
        fractions=fractions,
        fractions_face=fractions_face,
        A_avg=A_avg,
        A_avg_face=A_avg_face,
        Z_override=Z_override,
    )

  @pydantic.model_validator(mode='before')
  @classmethod
  def _conform_impurity_data(cls, data: dict[str, Any]) -> dict[str, Any]:
    """Ensures backward compatibility if infered that data in legacy format."""

    # Maps legacy inputs to the new API format and convert what would have been
    # TimeVaryingScalar inputs to an equivalent TimeVaryingArray.
    # TODO(b/434175938): Remove this once V1 API is deprecated.
    # This branch is hit when calling from the outer `PlasmaComposition`.
    if 'legacy' in data:
      del data['legacy']
      if 'species' in data and isinstance(data['species'], dict):
        new_species = {}
        for species, value in data['species'].items():
          new_species[species] = (
              torax_pydantic.TimeVaryingScalar.model_validate(value)
              .to_time_varying_array()
          )
        data['species'] = new_species

    # This branch is typically hit on using `update_fields`.
    if 'species' not in data and 'impurity_mode' not in data:
      if isinstance(data, dict):
        new_species = {}
        for species, value in data.items():
          new_species[species] = (
              torax_pydantic.TimeVaryingScalar.model_validate(value)
              .to_time_varying_array()
          )
      else:
        new_species = data
      return {'species': new_species, 'impurity_mode': _IMPURITY_MODE_FRACTIONS}

    return data



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Base classes for Neoclassical models."""
import dataclasses

import jax
from torax._src.neoclassical.bootstrap_current import base as bootstrap_current_base
from torax._src.neoclassical.conductivity import base as conductivity_base
from torax._src.neoclassical.transport import base as transport_base


@jax.tree_util.register_dataclass
@dataclasses.dataclass
class NeoclassicalModels:
  """Container for instantiated Neoclassical model objects."""

  conductivity: conductivity_base.ConductivityModel
  bootstrap_current: bootstrap_current_base.BootstrapCurrentModel
  transport: transport_base.NeoclassicalTransportModel

  def __hash__(self) -> int:
    return hash((self.bootstrap_current, self.conductivity, self.transport))

  def __eq__(self, other) -> bool:
    return (
        isinstance(other, NeoclassicalModels)
        and self.conductivity == other.conductivity
        and self.bootstrap_current == other.bootstrap_current
        and self.transport == other.transport
    )



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Code for getting the initial state for a simulation."""
import dataclasses

from absl import logging
import numpy as np
from torax._src import state
from torax._src.config import build_runtime_params
from torax._src.config import runtime_params_slice
from torax._src.core_profiles import initialization
from torax._src.geometry import geometry
from torax._src.geometry import geometry_provider as geometry_provider_lib
from torax._src.neoclassical.conductivity import base as conductivity_base
from torax._src.orchestration import sim_state
from torax._src.orchestration import step_function
from torax._src.output_tools import output
from torax._src.output_tools import post_processing
from torax._src.sources import source_profile_builders
from torax._src.torax_pydantic import file_restart as file_restart_pydantic_model
from torax._src.transport_model import transport_coefficients_builder
import xarray as xr


def get_initial_state_and_post_processed_outputs(
    t: float,
    runtime_params_provider: build_runtime_params.RuntimeParamsProvider,
    geometry_provider: geometry_provider_lib.GeometryProvider,
    step_fn: step_function.SimulationStepFn,
) -> tuple[sim_state.ToraxSimState, post_processing.PostProcessedOutputs]:
  """Returns the initial state and post processed outputs."""
  runtime_params_for_init, geo_for_init = (
      build_runtime_params.get_consistent_runtime_params_and_geometry(
          t=t,
          runtime_params_provider=runtime_params_provider,
          geometry_provider=geometry_provider,
      )
  )
  initial_state = _get_initial_state(
      runtime_params=runtime_params_for_init,
      geo=geo_for_init,
      step_fn=step_fn,
  )
  post_processed_outputs = post_processing.make_post_processed_outputs(
      initial_state, runtime_params_for_init
  )
  return initial_state, post_processed_outputs


def _get_initial_state(
    runtime_params: runtime_params_slice.RuntimeParams,
    geo: geometry.Geometry,
    step_fn: step_function.SimulationStepFn,
) -> sim_state.ToraxSimState:
  """Returns the initial state to be used by run_simulation()."""
  physics_models = step_fn.solver.physics_models
  initial_core_profiles = initialization.initial_core_profiles(
      runtime_params,
      geo,
      source_models=physics_models.source_models,
      neoclassical_models=physics_models.neoclassical_models,
  )
  # Populate the starting state with source profiles from the implicit sources
  # before starting the run-loop. The explicit source profiles will be computed
  # inside the loop and will be merged with these implicit source profiles.
  initial_core_sources = source_profile_builders.get_all_source_profiles(
      runtime_params=runtime_params,
      geo=geo,
      core_profiles=initial_core_profiles,
      source_models=physics_models.source_models,
      neoclassical_models=physics_models.neoclassical_models,
      conductivity=conductivity_base.Conductivity(
          sigma=initial_core_profiles.sigma,
          sigma_face=initial_core_profiles.sigma_face,
      ),
  )

  transport_coeffs = (
      transport_coefficients_builder.calculate_total_transport_coeffs(
          physics_models.pedestal_model,
          physics_models.transport_model,
          physics_models.neoclassical_models,
          runtime_params,
          geo,
          initial_core_profiles,
      )
  )

  return sim_state.ToraxSimState(
      t=np.array(runtime_params.numerics.t_initial),
      dt=np.zeros(()),
      core_profiles=initial_core_profiles,
      core_sources=initial_core_sources,
      core_transport=transport_coeffs,
      solver_numeric_outputs=state.SolverNumericOutputs(
          solver_error_state=0,
          outer_solver_iterations=0,
          inner_solver_iterations=0,
      ),
      geometry=geo,
  )


def get_initial_state_and_post_processed_outputs_from_file(
    t_initial: float,
    file_restart: file_restart_pydantic_model.FileRestart,
    runtime_params_provider: build_runtime_params.RuntimeParamsProvider,
    geometry_provider: geometry_provider_lib.GeometryProvider,
    step_fn: step_function.SimulationStepFn,
) -> tuple[sim_state.ToraxSimState, post_processing.PostProcessedOutputs]:
  """Returns the initial state and post processed outputs from a file."""
  data_tree = output.load_state_file(file_restart.filename)
  # Find the closest time in the given dataset.
  data_tree = data_tree.sel(time=file_restart.time, method='nearest')
  t_restart = data_tree.time.item()
  profiles_dataset = data_tree.children[output.PROFILES].dataset
  profiles_dataset = profiles_dataset.squeeze()
  if t_restart != t_initial:
    logging.warning(
        'Requested restart time %f not exactly available in state file %s.'
        ' Restarting from closest available time %f instead.',
        file_restart.time,
        file_restart.filename,
        t_restart,
    )

  runtime_params_for_init, geo_for_init = (
      build_runtime_params.get_consistent_runtime_params_and_geometry(
          t=t_initial,
          runtime_params_provider=runtime_params_provider,
          geometry_provider=geometry_provider,
      )
  )
  runtime_params_for_init, geo_for_init = (
      _override_initial_runtime_params_from_file(
          runtime_params_for_init,
          geo_for_init,
          t_restart,
          profiles_dataset,
      )
  )
  initial_state = _get_initial_state(
      runtime_params=runtime_params_for_init,
      geo=geo_for_init,
      step_fn=step_fn,
  )
  scalars_dataset = data_tree.children[output.SCALARS].dataset
  scalars_dataset = scalars_dataset.squeeze()
  post_processed_outputs = post_processing.make_post_processed_outputs(
      initial_state,
      runtime_params_for_init,
  )
  post_processed_outputs = dataclasses.replace(
      post_processed_outputs,
      E_fusion=scalars_dataset.data_vars['E_fusion'].to_numpy(),
      E_aux=scalars_dataset.data_vars['E_aux'].to_numpy(),
  )
  core_profiles = dataclasses.replace(
      initial_state.core_profiles,
      v_loop_lcfs=scalars_dataset.v_loop_lcfs.values,
  )
  numerics_dataset = data_tree.children[output.NUMERICS].dataset
  numerics_dataset = numerics_dataset.squeeze()
  sawtooth_crash = bool(numerics_dataset[output.SAWTOOTH_CRASH])
  outer_solver_iterations = int(
      numerics_dataset[output.OUTER_SOLVER_ITERATIONS]
  )
  inner_solver_iterations = int(
      numerics_dataset[output.INNER_SOLVER_ITERATIONS]
  )
  return (
      dataclasses.replace(
          initial_state,
          core_profiles=core_profiles,
          solver_numeric_outputs=state.SolverNumericOutputs(
              sawtooth_crash=sawtooth_crash,
              outer_solver_iterations=outer_solver_iterations,
              inner_solver_iterations=inner_solver_iterations,
          ),
      ),
      post_processed_outputs,
  )


def _override_initial_runtime_params_from_file(
    runtime_params: runtime_params_slice.RuntimeParams,
    geo: geometry.Geometry,
    t_restart: float,
    profiles_ds: xr.Dataset,
) -> tuple[runtime_params_slice.RuntimeParams, geometry.Geometry]:
  """Override parts of runtime params slice from state in a file."""
  # pylint: disable=invalid-name
  runtime_params.numerics.t_initial = t_restart
  runtime_params.profile_conditions.Ip = profiles_ds.data_vars[
      output.IP_PROFILE
  ].to_numpy()[-1]
  runtime_params.profile_conditions.T_e = (
      profiles_ds.data_vars[output.T_E]
      .sel(rho_norm=profiles_ds.coords[output.RHO_CELL_NORM])
      .to_numpy()
  )
  runtime_params.profile_conditions.T_e_right_bc = (
      profiles_ds.data_vars[output.T_E]
      .sel(rho_norm=profiles_ds.coords[output.RHO_FACE_NORM][-1])
      .to_numpy()
  )
  runtime_params.profile_conditions.T_i = (
      profiles_ds.data_vars[output.T_I]
      .sel(rho_norm=profiles_ds.coords[output.RHO_CELL_NORM])
      .to_numpy()
  )
  runtime_params.profile_conditions.T_i_right_bc = (
      profiles_ds.data_vars[output.T_I]
      .sel(rho_norm=profiles_ds.coords[output.RHO_FACE_NORM][-1])
      .to_numpy()
  )
  # Density in output is in m^-3.
  runtime_params.profile_conditions.n_e = (
      profiles_ds.data_vars[output.N_E]
      .sel(rho_norm=profiles_ds.coords[output.RHO_CELL_NORM])
      .to_numpy()
  )
  runtime_params.profile_conditions.n_e_right_bc = (
      profiles_ds.data_vars[output.N_E]
      .sel(rho_norm=profiles_ds.coords[output.RHO_FACE_NORM][-1])
      .to_numpy()
  )
  runtime_params.profile_conditions.psi = (
      profiles_ds.data_vars[output.PSI]
      .sel(rho_norm=profiles_ds.coords[output.RHO_CELL_NORM])
      .to_numpy()
  )
  # When loading from file we want ne not to have transformations.
  # Both ne and the boundary condition are given in absolute values (not fGW).
  # Additionally we want to avoid normalizing to nbar.
  runtime_params.profile_conditions.n_e_right_bc_is_fGW = False
  runtime_params.profile_conditions.n_e_nbar_is_fGW = False
  runtime_params.profile_conditions.normalize_n_e_to_nbar = False
  runtime_params.profile_conditions.n_e_right_bc_is_absolute = True

  return runtime_params, geo



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The FixedTimeStepCalculator class.

Steps through time using a constant time step.
"""

import jax
from jax import numpy as jnp
from torax._src import state as state_module
from torax._src.config import runtime_params_slice
from torax._src.geometry import geometry
from torax._src.time_step_calculator import time_step_calculator


# TODO(b/337844885). Make this a time dependent calculator.
class FixedTimeStepCalculator(time_step_calculator.TimeStepCalculator):
  """TimeStepCalculator based on constant time steps."""

  def _next_dt(
      self,
      runtime_params: runtime_params_slice.RuntimeParams,
      geo: geometry.Geometry,
      core_profiles: state_module.CoreProfiles,
      core_transport: state_module.CoreTransport,
  ) -> jax.Array:
    """Returns the fixed time step duration."""
    return jnp.array(runtime_params.numerics.fixed_dt)

  def __eq__(self, other) -> bool:
    return isinstance(other, type(self))

  def __hash__(self) -> int:
    return hash(type(self))



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""JITted run_loop for iterating over the simulation step function."""
import functools

import chex
import jax
import jax.numpy as jnp
from torax._src import jax_utils
from torax._src import state
from torax._src.config import build_runtime_params
from torax._src.orchestration import sim_state
from torax._src.orchestration import step_function
from torax._src.output_tools import post_processing


@functools.partial(jax_utils.jit, static_argnames=['max_steps'])
def run_loop_jit(
    initial_state: sim_state.ToraxSimState,
    initial_post_processed_outputs: post_processing.PostProcessedOutputs,
    step_fn: step_function.SimulationStepFn,
    max_steps: int,
) -> tuple[
    sim_state.ToraxSimState, post_processing.PostProcessedOutputs, chex.Numeric
]:
  """Runs the simulation loop under jax.jit."""
  # Some of the runtime params are not time-dependent, so we can get them once
  # before the loop.
  initial_runtime_params = step_fn.runtime_params_provider(initial_state.t)
  time_step_calculator_params = initial_runtime_params.time_step_calculator

  # Pre-allocate history buffers
  states_history = jax.tree_util.tree_map(
      lambda x: jnp.zeros((max_steps + 1,) + x.shape, dtype=x.dtype),
      initial_state,
  )
  post_processed_outputs_history = jax.tree_util.tree_map(
      lambda x: jnp.zeros((max_steps + 1,) + x.shape, dtype=x.dtype),
      initial_post_processed_outputs,
  )

  # Store initial state
  states_history = jax.tree_util.tree_map(
      lambda hist, val: hist.at[0].set(val), states_history, initial_state
  )
  post_processed_outputs_history = jax.tree_util.tree_map(
      lambda hist, val: hist.at[0].set(val),
      post_processed_outputs_history,
      initial_post_processed_outputs,
  )

  def _cond_fun(inputs):
    i, current_state, _, _, _ = inputs
    not_done = step_fn.time_step_calculator.not_done(
        current_state.t,
        step_fn.runtime_params_provider.numerics.t_final,
        time_step_calculator_params,
    )
    return jnp.logical_and(i < max_steps, not_done)

  def _step_fn(inputs):
    (
        i,
        previous_state,
        previous_post_processed_outputs,
        states_hist,
        post_processed_outputs_hist,
    ) = inputs
    current_state, post_processed_outputs = step_fn(
        previous_state, previous_post_processed_outputs
    )
    states_hist = jax.tree_util.tree_map(
        lambda hist, val: hist.at[i + 1].set(val), states_hist, current_state
    )
    post_processed_outputs_hist = jax.tree_util.tree_map(
        lambda hist, val: hist.at[i + 1].set(val),
        post_processed_outputs_hist,
        post_processed_outputs,
    )
    return (
        i + 1,
        current_state,
        post_processed_outputs,
        states_hist,
        post_processed_outputs_hist,
    )

  final_i, _, _, states_history, post_processed_outputs_history = (
      jax.lax.while_loop(
          _cond_fun,
          _step_fn,
          (
              0,
              initial_state,
              initial_post_processed_outputs,
              states_history,
              post_processed_outputs_history,
          ),
      )
  )

  return states_history, post_processed_outputs_history, final_i


def _unstack_pytree_history(
    states_history,
    post_processed_outputs_history,
    final_i,
):
  """Unstacks stacked JIT output into a list of unstacked outputs.

  Args:
    states_history: A PyTree where each leaf is a JAX array with shape
      (max_steps + 1, ...) representing the history of that component over time.
    post_processed_outputs_history: A PyTree where each leaf is a JAX array with
      shape (max_steps + 1, ...) representing the history of that component over
      time.
    final_i: The actual number of steps taken in the simulation.

  Returns:
    A list of PyTrees, where each element of the list corresponds to
    a time step [0, max_steps]. Each element of the list has the same
    structure and leaf types as the original initial_state.
  """
  states_history_list = []
  post_processed_outputs_history_list = []
  for time_index in range(final_i + 1):
    # Use tree_map to slice each leaf array at the current time step `i`
    current_state = jax.tree_util.tree_map(
        lambda x, i=time_index: x[i], states_history
    )
    states_history_list.append(current_state)
    post_processed_output = jax.tree_util.tree_map(
        lambda x, i=time_index: x[i], post_processed_outputs_history
    )
    post_processed_outputs_history_list.append(post_processed_output)
  return states_history_list, post_processed_outputs_history_list


def run_loop(
    runtime_params_provider: build_runtime_params.RuntimeParamsProvider,
    initial_state: sim_state.ToraxSimState,
    initial_post_processed_outputs: post_processing.PostProcessedOutputs,
    step_fn: step_function.SimulationStepFn,
    log_timestep_info: bool = False,
    progress_bar: bool = True,
) -> tuple[
    tuple[sim_state.ToraxSimState, ...],
    tuple[post_processing.PostProcessedOutputs, ...],
    state.SimError,
]:
  """Version of torax._src.orchestration.run_loop that loops with jax.jit.

  This function must be run with TORAX_EXPERIMENTAL_COMPILE=True.
  Unlike the `run_loop` function, This does not support logging or progress bar.

  Args:
    runtime_params_provider: Provides a RuntimeParams to use as input for each
      time step.
    initial_state: The starting state of the simulation. This includes both the
      state variables which the solver.Solver will evolve (like ion temp, psi,
      etc.) as well as other states that need to be be tracked, like time.
    initial_post_processed_outputs: The post-processed outputs at the start of
      the simulation. This is used to calculate cumulative quantities.
    step_fn: Callable which takes in ToraxSimState and outputs the ToraxSimState
      after one timestep. Note that step_fn determines dt (how long the timestep
      is). The state_history that run_simulation() outputs comes from these
      ToraxSimState objects.
    log_timestep_info: If True, logs basic timestep info, like time, dt, on
      every step.
    progress_bar: If True, displays a progress bar.

  Returns:
    A tuple of:
      - the simulation history, consisting of a tuple of ToraxSimState objects,
        one for each time step. There are N+1 objects returned, where N is the
        number of simulation steps taken. The first object in the tuple is for
        the initial state. If the sim error state is 1, then a trunctated
        simulation history is returned up until the last valid timestep.
      - the post-processed outputs history, consisting of a tuple of
        PostProcessedOutputs objects, one for each time step. There are N+1
        objects returned, where N is the number of simulation steps taken. The
        first object in the tuple is for the initial state. If the sim error
        state is 1, then a trunctated simulation history is returned up until
        the last valid timestep.
      - The sim error state.
  """
  del log_timestep_info, progress_bar
  numerics = runtime_params_provider.numerics
  max_steps = int(
      ((numerics.t_final - numerics.t_initial) / numerics.min_dt) / 1e5
  )
  states_history, post_processed_outputs_history, final_i = run_loop_jit(
      initial_state,
      initial_post_processed_outputs,
      step_fn,
      max_steps,
  )
  unstacked_states, unstacked_post_processed_outputs = _unstack_pytree_history(
      states_history, post_processed_outputs_history, final_i
  )
  sim_error = step_function.check_for_errors(
      runtime_params_provider.numerics,
      unstacked_states[-1],
      unstacked_post_processed_outputs[-1],
  )
  return (
      tuple(unstacked_states),
      tuple(unstacked_post_processed_outputs),
      sim_error,
  )



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Source/sink profiles for all the sources in TORAX."""
import dataclasses
from typing import Literal

import jax
import jax.numpy as jnp
from torax._src import constants
from torax._src.geometry import geometry
from torax._src.neoclassical.bootstrap_current import base as bootstrap_current_base
import typing_extensions

# pylint: disable=invalid-name


@jax.tree_util.register_dataclass
@dataclasses.dataclass(frozen=True)
class QeiInfo:
  """Represents the source values coming from a QeiSource."""

  qei_coef: jax.Array
  implicit_ii: jax.Array
  explicit_i: jax.Array
  implicit_ee: jax.Array
  explicit_e: jax.Array
  implicit_ie: jax.Array
  implicit_ei: jax.Array

  @classmethod
  def zeros(cls, geo: geometry.Geometry) -> typing_extensions.Self:
    return QeiInfo(
        qei_coef=jnp.zeros_like(geo.rho),
        implicit_ii=jnp.zeros_like(geo.rho),
        explicit_i=jnp.zeros_like(geo.rho),
        implicit_ee=jnp.zeros_like(geo.rho),
        explicit_e=jnp.zeros_like(geo.rho),
        implicit_ie=jnp.zeros_like(geo.rho),
        implicit_ei=jnp.zeros_like(geo.rho),
    )


@jax.tree_util.register_dataclass
@dataclasses.dataclass(frozen=True)
class SourceProfiles:
  """Collection of profiles for all sources in TORAX.

  Most profiles are stored in the attributes relating to the core profile they
  affect, but special-case profiles `j_bootstrap` and `qei` are pulled out into
  their own attributes as these sources need to be treated differently (though
  they could still be set to zero using appropriate runtime params).

  This dataclass is inspired by the IMAS `core_sources` IDS. It is not a 1:1
  mapping to that schema, but it contains similar profiles as you'd expect in
  that IDS.
  """

  # Special-case profiles.
  bootstrap_current: bootstrap_current_base.BootstrapCurrent
  qei: QeiInfo
  # Other profiles organised by the affected core profile. These are the
  # profiles that are used to compute the core profile equations.
  # The form is a dict of jax.Arrays, keyed by the name of the source. The array
  # is the profile on the cell grid from that source for that core profile.
  # For sources that affect multiple core profiles, they will have an entry for
  # each core profile they affect.
  T_e: dict[str, jax.Array] = dataclasses.field(default_factory=dict)
  T_i: dict[str, jax.Array] = dataclasses.field(default_factory=dict)
  n_e: dict[str, jax.Array] = dataclasses.field(default_factory=dict)
  psi: dict[str, jax.Array] = dataclasses.field(default_factory=dict)

  # This function can be jitted if source_models is a static argument. However,
  # in our tests, jitting this function actually slightly slows down runs, so
  # this is left as pure python.
  @classmethod
  def merge(
      cls,
      explicit_source_profiles: typing_extensions.Self,
      implicit_source_profiles: typing_extensions.Self,
  ) -> typing_extensions.Self:
    """Returns a SourceProfiles that merges the input profiles.

    Sources can either be explicit or implicit. The explicit_source_profiles
    contain the profiles for all source models that are set to explicit, and it
    contains profiles with all zeros for any implicit source. The opposite holds
    for the implicit_source_profiles.

    This function adds the two dictionaries of profiles and returns a single
    SourceProfiles that includes both.

    Args:
      explicit_source_profiles: Profiles from explicit source models. This
        SourceProfiles dict will include keys for both the explicit and implicit
        sources, but only the explicit sources will have non-zero profiles. See
        source.py and runtime_params.py for more info on explicit vs. implicit.
      implicit_source_profiles: Profiles from implicit source models. This
        SourceProfiles dict will include keys for both the explicit and implicit
        sources, but only the implicit sources will have non-zero profiles. See
        source.py and runtime_params.py for more info on explicit vs. implicit.

    Returns:
      A SourceProfiles with non-zero profiles for all sources, both explicit and
      implicit (assuming the source model outputted a non-zero profile).

    """
    sum_profiles = lambda a, b: a + b
    return jax.tree_util.tree_map(
        sum_profiles, explicit_source_profiles, implicit_source_profiles
    )

  def total_psi_sources(self, geo: geometry.Geometry) -> jax.Array:
    total = self.bootstrap_current.j_bootstrap
    total += sum(self.psi.values())
    mu0 = constants.CONSTANTS.mu0
    prefactor = 8 * geo.vpr * jnp.pi**2 * geo.B_0 * mu0 * geo.Phi_b / geo.F**2
    return -total * prefactor

  def total_sources(
      self,
      source_type: Literal['n_e', 'T_i', 'T_e'],
      geo: geometry.Geometry,
  ) -> jax.Array:
    source: dict[str, jax.Array] = getattr(self, source_type)
    total = sum(source.values())
    return total * geo.vpr



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Zeros model for neoclassical transport."""
from typing import Annotated, Literal

import jax.numpy as jnp
from torax._src import state
from torax._src.config import runtime_params_slice
from torax._src.geometry import geometry as geometry_lib
from torax._src.neoclassical.transport import base
from torax._src.torax_pydantic import torax_pydantic
from typing_extensions import override


class ZerosModel(base.NeoclassicalTransportModel):
  """Zeros model for neoclassical transport."""

  @override
  def _call_implementation(
      self,
      runtime_params: runtime_params_slice.RuntimeParams,
      geometry: geometry_lib.Geometry,
      core_profiles: state.CoreProfiles,
  ) -> base.NeoclassicalTransport:
    """Calculates neoclassical transport."""
    return base.NeoclassicalTransport(
        chi_neo_i=jnp.zeros_like(geometry.rho_face),
        chi_neo_e=jnp.zeros_like(geometry.rho_face),
        D_neo_e=jnp.zeros_like(geometry.rho_face),
        V_neo_e=jnp.zeros_like(geometry.rho_face),
        V_neo_ware_e=jnp.zeros_like(geometry.rho_face),
    )

  def __eq__(self, other) -> bool:
    return isinstance(other, self.__class__)

  def __hash__(self) -> int:
    return hash(self.__class__)


class ZerosModelConfig(base.NeoclassicalTransportModelConfig):
  """Config for the Zeros model implementation of neoclassical transport."""

  model_name: Annotated[Literal['zeros'], torax_pydantic.JAX_STATIC] = 'zeros'

  def build_model(self) -> ZerosModel:
    return ZerosModel()



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The Solver class.

Abstract base class defining updates to State.
"""

import abc
import functools

import jax
from torax._src import jax_utils
from torax._src import physics_models as physics_models_lib
from torax._src import state
from torax._src.config import runtime_params_slice
from torax._src.fvm import cell_variable
from torax._src.geometry import geometry
from torax._src.sources import source_profiles
import typing_extensions


class Solver(abc.ABC):
  """Solves for a single time steps update to State.

  Attributes:
    physics_models: Physics models.
  """

  def __init__(
      self,
      physics_models: physics_models_lib.PhysicsModels,
  ):
    self.physics_models = physics_models

  def __hash__(self) -> int:
    return hash(self.physics_models)

  def __eq__(self, other: typing_extensions.Self) -> bool:
    return self.physics_models == other.physics_models

  @functools.partial(
      jax_utils.jit,
      static_argnames=[
          'self',
      ],
  )
  def __call__(
      self,
      t: jax.Array,
      dt: jax.Array,
      runtime_params_t: runtime_params_slice.RuntimeParams,
      runtime_params_t_plus_dt: runtime_params_slice.RuntimeParams,
      geo_t: geometry.Geometry,
      geo_t_plus_dt: geometry.Geometry,
      core_profiles_t: state.CoreProfiles,
      core_profiles_t_plus_dt: state.CoreProfiles,
      explicit_source_profiles: source_profiles.SourceProfiles,
  ) -> tuple[
      tuple[cell_variable.CellVariable, ...],
      state.SolverNumericOutputs,
  ]:
    """Applies a time step update.

    Args:
      t: Time.
      dt: Time step duration.
      runtime_params_t: Runtime parameters for time t (the start time of the
        step). These runtime params can change from step to step without
        triggering a recompilation.
      runtime_params_t_plus_dt: Runtime parameters for time t + dt, used for
        implicit calculations in the solver.
      geo_t: Geometry of the torus at time t.
      geo_t_plus_dt: Geometry of the torus at time t + dt.
      core_profiles_t: Core plasma profiles at the beginning of the time step.
      core_profiles_t_plus_dt: Core plasma profiles which contain all available
        prescribed quantities at the end of the time step. This includes
        evolving boundary conditions and prescribed time-dependent profiles that
        are not being evolved by the PDE system.
      explicit_source_profiles: Source profiles of all explicit sources (as
        configured by the input params). All implicit source's profiles will be
        set to 0 in this object. These explicit source profiles were calculated
        either based on the original core profiles at the start of the time step
        or were independent of the core profiles. Because they were calculated
        outside the possibly-JAX-jitted solver logic, they can be calculated in
        non-JAX-friendly ways.

    Returns:
      x_new: Tuple containing new cell-grid values of the evolving variables.
      solver_numeric_output: Error and solver iteration info.
    """

    # This base class method can be completely overridden by a subclass, but
    # most can make use of the boilerplate here and just implement `_x_new`.

    # Don't call solver functions on an empty list
    if runtime_params_t.numerics.evolving_names:
      (
          x_new,
          solver_numeric_output,
      ) = self._x_new(
          dt=dt,
          runtime_params_t=runtime_params_t,
          runtime_params_t_plus_dt=runtime_params_t_plus_dt,
          geo_t=geo_t,
          geo_t_plus_dt=geo_t_plus_dt,
          core_profiles_t=core_profiles_t,
          core_profiles_t_plus_dt=core_profiles_t_plus_dt,
          explicit_source_profiles=explicit_source_profiles,
          evolving_names=runtime_params_t.numerics.evolving_names,
      )
    else:
      x_new = tuple()
      solver_numeric_output = state.SolverNumericOutputs()

    return (
        x_new,
        solver_numeric_output,
    )

  def _x_new(
      self,
      dt: jax.Array,
      runtime_params_t: runtime_params_slice.RuntimeParams,
      runtime_params_t_plus_dt: runtime_params_slice.RuntimeParams,
      geo_t: geometry.Geometry,
      geo_t_plus_dt: geometry.Geometry,
      core_profiles_t: state.CoreProfiles,
      core_profiles_t_plus_dt: state.CoreProfiles,
      explicit_source_profiles: source_profiles.SourceProfiles,
      evolving_names: tuple[str, ...],
  ) -> tuple[
      tuple[cell_variable.CellVariable, ...],
      state.SolverNumericOutputs,
  ]:
    """Calculates new values of the changing variables.

    Subclasses must either implement `_x_new` so that `Solver.__call__`
    will work, or implement a different `__call__`.

    Args:
      dt: Time step duration.
      runtime_params_t: Runtime parameters for time t (the start time of the
        step). These runtime params can change from step to step without
        triggering a recompilation.
      runtime_params_t_plus_dt: Runtime parameters for time t + dt, used for
        implicit calculations in the solver.
      geo_t: Geometry of the torus for time t.
      geo_t_plus_dt: Geometry of the torus for time t + dt.
      core_profiles_t: Core plasma profiles at the beginning of the time step.
      core_profiles_t_plus_dt: Core plasma profiles which contain all available
        prescribed quantities at the end of the time step. This includes
        evolving boundary conditions and prescribed time-dependent profiles that
        are not being evolved by the PDE system.
      explicit_source_profiles: see the docstring of __call__
      evolving_names: The names of core_profiles variables that should evolve.

    Returns:
      x_new: The values of the evolving variables at time t + dt.
      solver_numeric_output: Error and solver iteration info.
    """

    raise NotImplementedError(
        f'{type(self)} must implement `_x_new` or '
        'implement a different `__call__` that does not'
        ' need `_x_new`.'
    )



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Code to build the combined transport coefficients for a simulation."""
import dataclasses
import functools

from torax._src import jax_utils
from torax._src import state
from torax._src.config import runtime_params_slice
from torax._src.geometry import geometry
from torax._src.neoclassical import neoclassical_models as neoclassical_models_lib
from torax._src.pedestal_model import pedestal_model as pedestal_model_lib
from torax._src.transport_model import transport_model as transport_model_lib


@functools.partial(jax_utils.jit, static_argnums=(0, 1, 2))
def calculate_total_transport_coeffs(
    pedestal_model: pedestal_model_lib.PedestalModel,
    transport_model: transport_model_lib.TransportModel,
    neoclassical_models: neoclassical_models_lib.NeoclassicalModels,
    runtime_params: runtime_params_slice.RuntimeParams,
    geo: geometry.Geometry,
    core_profiles: state.CoreProfiles,
) -> state.CoreTransport:
  """Calculates the transport coefficients."""
  pedestal_model_output = pedestal_model(runtime_params, geo, core_profiles)
  turbulent_transport = transport_model(
      runtime_params=runtime_params,
      geo=geo,
      core_profiles=core_profiles,
      pedestal_model_output=pedestal_model_output,
  )
  neoclassical_transport_coeffs = neoclassical_models.transport(
      runtime_params,
      geo,
      core_profiles,
  )

  return state.CoreTransport(
      **dataclasses.asdict(turbulent_transport),
      **dataclasses.asdict(neoclassical_transport_coeffs),
  )



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A container for all physics models."""
import dataclasses

import jax
from torax._src.mhd import base as mhd_model_lib
from torax._src.neoclassical import neoclassical_models as neoclassical_models_lib
from torax._src.pedestal_model import pedestal_model as pedestal_model_lib
from torax._src.sources import source_models as source_models_lib
from torax._src.transport_model import transport_model as transport_model_lib


@jax.tree_util.register_dataclass
@dataclasses.dataclass(frozen=True)
class PhysicsModels:
  """A container for all physics models."""

  source_models: source_models_lib.SourceModels = dataclasses.field(
      metadata=dict(static=True)
  )
  transport_model: transport_model_lib.TransportModel = dataclasses.field(
      metadata=dict(static=True)
  )
  pedestal_model: pedestal_model_lib.PedestalModel = dataclasses.field(
      metadata=dict(static=True)
  )
  neoclassical_models: neoclassical_models_lib.NeoclassicalModels = (
      dataclasses.field(metadata=dict(static=True))
  )
  mhd_models: mhd_model_lib.MHDModels = dataclasses.field(
      metadata=dict(static=True)
  )



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A basic version of the pedestal model that uses direct specification."""
import dataclasses

import jax
from jax import numpy as jnp
from torax._src import array_typing
from torax._src import state
from torax._src.config import runtime_params_slice
from torax._src.geometry import geometry
from torax._src.pedestal_model import pedestal_model
from torax._src.pedestal_model import runtime_params as runtime_params_lib
from typing_extensions import override


# pylint: disable=invalid-name
@jax.tree_util.register_dataclass
@dataclasses.dataclass(frozen=True)
class RuntimeParams(runtime_params_lib.RuntimeParams):
  """Runtime params for the SetTemperatureDensityPedestalModel."""

  n_e_ped: array_typing.FloatScalar
  T_i_ped: array_typing.FloatScalar
  T_e_ped: array_typing.FloatScalar
  rho_norm_ped_top: array_typing.FloatScalar
  n_e_ped_is_fGW: array_typing.BoolScalar


class SetTemperatureDensityPedestalModel(pedestal_model.PedestalModel):
  """A basic version of the pedestal model that uses direct specification."""

  def __init__(
      self,
  ):
    super().__init__()
    self._frozen = True

  @override
  def _call_implementation(
      self,
      runtime_params: runtime_params_slice.RuntimeParams,
      geo: geometry.Geometry,
      core_profiles: state.CoreProfiles,
  ) -> pedestal_model.PedestalModelOutput:
    pedestal_params = runtime_params.pedestal
    assert isinstance(pedestal_params, RuntimeParams)
    nGW = (
        runtime_params.profile_conditions.Ip
        / 1e6  # Convert to MA.
        / (jnp.pi * geo.a_minor**2)
        * 1e20
    )
    # Calculate n_e_ped in m^-3.
    n_e_ped = jnp.where(
        pedestal_params.n_e_ped_is_fGW,
        pedestal_params.n_e_ped * nGW,
        pedestal_params.n_e_ped,
    )
    return pedestal_model.PedestalModelOutput(
        n_e_ped=n_e_ped,
        T_i_ped=pedestal_params.T_i_ped,
        T_e_ped=pedestal_params.T_e_ped,
        rho_norm_ped_top=pedestal_params.rho_norm_ped_top,
        rho_norm_ped_top_idx=jnp.abs(
            geo.rho_norm - pedestal_params.rho_norm_ped_top
        ).argmin(),
    )

  def __hash__(self) -> int:
    return hash('SetTemperatureDensityPedestalModel')

  def __eq__(self, other) -> bool:
    return isinstance(other, SetTemperatureDensityPedestalModel)



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Zeros model for bootstrap current."""
from typing import Annotated, Literal

import jax.numpy as jnp
from torax._src import state
from torax._src.config import runtime_params_slice
from torax._src.geometry import geometry as geometry_lib
from torax._src.neoclassical.bootstrap_current import base
from torax._src.neoclassical.bootstrap_current import runtime_params as bootstrap_runtime_params
from torax._src.torax_pydantic import torax_pydantic


class ZerosModel(base.BootstrapCurrentModel):
  """Zeros model for bootstrap current."""

  def calculate_bootstrap_current(
      self,
      runtime_params: runtime_params_slice.RuntimeParams,
      geometry: geometry_lib.Geometry,
      core_profiles: state.CoreProfiles,
  ) -> base.BootstrapCurrent:
    """Calculates bootstrap current."""
    return base.BootstrapCurrent(
        j_bootstrap=jnp.zeros_like(geometry.rho),
        j_bootstrap_face=jnp.zeros_like(geometry.rho_face),
    )

  def __eq__(self, other) -> bool:
    return isinstance(other, self.__class__)

  def __hash__(self) -> int:
    return hash(self.__class__)


class ZerosModelConfig(base.BootstrapCurrentModelConfig):
  """Config for the Zeros model implementation of bootstrap current."""

  model_name: Annotated[Literal['zeros'], torax_pydantic.JAX_STATIC] = 'zeros'

  def build_runtime_params(self) -> bootstrap_runtime_params.RuntimeParams:
    return bootstrap_runtime_params.RuntimeParams()

  def build_model(self) -> ZerosModel:
    return ZerosModel()



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Simple trigger model for sawteeth."""

import dataclasses
from typing import Annotated, Literal

import chex
import jax
from jax import numpy as jnp
from torax._src import array_typing
from torax._src import constants
from torax._src import state
from torax._src.config import runtime_params_slice
from torax._src.geometry import geometry
from torax._src.mhd.sawtooth import runtime_params as sawtooth_runtime_params
from torax._src.mhd.sawtooth import trigger_base
from torax._src.torax_pydantic import torax_pydantic


class SimpleTrigger(trigger_base.TriggerModel):
  """Simple trigger model."""

  def __call__(
      self,
      runtime_params: runtime_params_slice.RuntimeParams,
      geo: geometry.Geometry,
      core_profiles: state.CoreProfiles,
  ) -> tuple[array_typing.BoolScalar, array_typing.FloatScalar]:
    """Checks if the simple critical shear condition for a sawtooth is met.

    Args:
      runtime_params: Runtime parameters.
      geo: Geometry object.
      core_profiles: Core plasma profiles.

    Returns:
      tuple of (True if sawtooth crash is triggered, False otherwise,
        radius of q=1 surface (set to 0.0 if no surface exists))
    """
    # Extract necessary parameters
    sawtooth_params = runtime_params.mhd.sawtooth
    assert isinstance(sawtooth_params, sawtooth_runtime_params.RuntimeParams)
    assert isinstance(sawtooth_params.trigger_params, RuntimeParams)
    minimum_radius = sawtooth_params.trigger_params.minimum_radius
    s_critical = sawtooth_params.trigger_params.s_critical

    q_face = core_profiles.q_face
    s_face = core_profiles.s_face
    rho_face_norm = geo.rho_face_norm
    eps = constants.CONSTANTS.eps

    # Find the rightmost location where q crosses 1.0.
    # Allows for non-monotonic q profiles.

    # Calculate interpolated rho_q1 for all intervals
    q1 = q_face[:-1]
    q2 = q_face[1:]
    rho_norm1 = rho_face_norm[:-1]
    rho_norm2 = rho_face_norm[1:]
    # Handle division by zero if q1 == q2
    dq = jnp.where(jnp.abs(q2 - q1) > eps, q2 - q1, eps)
    # Calculate value of rho for q=1 by linearization of the q-profile
    # in the interval. If the calculated rho is within the interval, then the
    # q=1 surface exists for that interval.
    potential_rho_norm_q1 = (
        rho_norm1 + (rho_norm2 - rho_norm1) * (1.0 - q1) / dq
    )
    valid_interval_mask = jnp.logical_and(
        potential_rho_norm_q1 >= rho_norm1, potential_rho_norm_q1 <= rho_norm2
    )

    # Mask out intervals where there isn't actually a crossing
    # We set invalid rhos to 0.0, as rho_q1 must be > 0 for a valid surface
    valid_rho_norm_q1 = jnp.where(
        valid_interval_mask, potential_rho_norm_q1, 0.0
    )

    # The rightmost crossing corresponds to the maximum valid rho_q1
    # For no crossings, rho_q1 = 0.0 which will not trigger the sawtooth.
    rho_norm_q1 = jnp.max(valid_rho_norm_q1)
    s_at_q1 = jnp.interp(rho_norm_q1, rho_face_norm, s_face)

    # Check trigger conditions
    rho_norm_above_minimum = rho_norm_q1 > minimum_radius
    s_above_critical = s_at_q1 > s_critical

    return (
        jnp.logical_and(rho_norm_above_minimum, s_above_critical),
        rho_norm_q1,
    )

  def __hash__(self) -> int:
    return hash(self.__class__.__name__)

  def __eq__(self, other: object) -> bool:
    return isinstance(other, SimpleTrigger)


@jax.tree_util.register_dataclass
@dataclasses.dataclass(frozen=True)
class RuntimeParams(sawtooth_runtime_params.TriggerRuntimeParams):
  """Runtime params for simple trigger model.

  Attributes:
    s_critical: Critical shear value at q=1 for sawtooth triggering.
  """

  s_critical: array_typing.FloatScalar


class SimpleTriggerConfig(trigger_base.TriggerConfig):
  """Pydantic model for simple trigger configuration.

  Attributes:
    s_critical: Critical shear value.
  """

  s_critical: torax_pydantic.TimeVaryingScalar = (
      torax_pydantic.ValidatedDefault(0.1)
  )
  model_name: Annotated[Literal['simple'], torax_pydantic.JAX_STATIC] = 'simple'

  def build_runtime_params(
      self,
      t: chex.Numeric,
  ) -> RuntimeParams:
    base_kwargs = dataclasses.asdict(super().build_runtime_params(t))
    return RuntimeParams(
        **base_kwargs,
        s_critical=self.s_critical.get_value(t),
    )

  def build_trigger_model(self) -> SimpleTrigger:
    return SimpleTrigger()



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pydantic model for sawtooth configuration."""
import chex
import pydantic
from torax._src.mhd.sawtooth import runtime_params as sawtooth_runtime_params
from torax._src.mhd.sawtooth import sawtooth_models
from torax._src.mhd.sawtooth import simple_redistribution
from torax._src.mhd.sawtooth import simple_trigger
from torax._src.torax_pydantic import torax_pydantic


class SawtoothConfig(torax_pydantic.BaseModelFrozen):
  """Pydantic model for sawtooth configuration.

  Attributes:
    trigger_model: Configuration for the trigger model.
    redistribution_model: Configuration for the redistribution model.
    crash_step_duration: Sawteeth crash period for extra timestep generated.
  """

  trigger_model: simple_trigger.SimpleTriggerConfig = pydantic.Field(
      discriminator='model_name'
  )

  redistribution_model: simple_redistribution.SimpleRedistributionConfig = (
      pydantic.Field(discriminator='model_name')
  )

  crash_step_duration: torax_pydantic.Second = 1e-3

  def build_models(self):
    return sawtooth_models.SawtoothModels(
        trigger_model=self.trigger_model.build_trigger_model(),
        redistribution_model=self.redistribution_model.build_redistribution_model(),
    )

  def build_runtime_params(
      self, t: chex.Numeric
  ) -> sawtooth_runtime_params.RuntimeParams:
    return sawtooth_runtime_params.RuntimeParams(
        crash_step_duration=self.crash_step_duration,
        trigger_params=self.trigger_model.build_runtime_params(t),
        redistribution_params=self.redistribution_model.build_runtime_params(t),
    )



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""External current source profile."""
import dataclasses
from typing import Annotated, ClassVar, Literal

import chex
import jax
from jax import numpy as jnp
from torax._src import array_typing
from torax._src import math_utils
from torax._src import state
from torax._src.config import runtime_params_slice
from torax._src.geometry import geometry
from torax._src.neoclassical.conductivity import base as conductivity_base
from torax._src.sources import base as source_base
from torax._src.sources import runtime_params as runtime_params_lib
from torax._src.sources import source
from torax._src.sources import source_profiles
from torax._src.torax_pydantic import torax_pydantic

# Default value for the model function to be used for the generic current
# source. This is also used as an identifier for the model function in
# the default source config for Pydantic to "discriminate" against.
DEFAULT_MODEL_FUNCTION_NAME: str = 'gaussian'


# pylint: disable=invalid-name
@jax.tree_util.register_dataclass
@dataclasses.dataclass(frozen=True)
class RuntimeParams(runtime_params_lib.RuntimeParams):
  """Runtime parameters for the external current source."""

  I_generic: array_typing.FloatScalar
  fraction_of_total_current: array_typing.FloatScalar
  gaussian_width: array_typing.FloatScalar
  gaussian_location: array_typing.FloatScalar
  use_absolute_current: bool


def calculate_generic_current(
    runtime_params: runtime_params_slice.RuntimeParams,
    geo: geometry.Geometry,
    source_name: str,
    unused_state: state.CoreProfiles,
    unused_calculated_source_profiles: source_profiles.SourceProfiles | None,
    unused_conductivity: conductivity_base.Conductivity | None,
) -> tuple[array_typing.FloatVectorCell, ...]:
  """Calculates the external current density profiles on the cell grid."""
  source_params = runtime_params.sources[source_name]
  # pytype: enable=name-error
  assert isinstance(source_params, RuntimeParams)
  I_generic = _calculate_I_generic(
      runtime_params,
      source_params,
  )
  # form of external current on cell grid
  generic_current_form = jnp.exp(
      -((geo.rho_norm - source_params.gaussian_location) ** 2)
      / (2 * source_params.gaussian_width**2)
  )

  Cext = I_generic / math_utils.area_integration(generic_current_form, geo)
  generic_current_profile = Cext * generic_current_form
  return (generic_current_profile,)


def _calculate_I_generic(
    runtime_params: runtime_params_slice.RuntimeParams,
    source_params: RuntimeParams,
) -> chex.Numeric:
  """Calculates the total value of external current."""
  return jnp.where(
      source_params.use_absolute_current,
      source_params.I_generic,
      (
          runtime_params.profile_conditions.Ip
          * source_params.fraction_of_total_current
      ),
  )


@dataclasses.dataclass(kw_only=True, frozen=True, eq=True)
class GenericCurrentSource(source.Source):
  """A generic current density source profile."""

  SOURCE_NAME: ClassVar[str] = 'generic_current'
  model_func: source.SourceProfileFunction = calculate_generic_current

  @property
  def source_name(self) -> str:
    return self.SOURCE_NAME

  @property
  def affected_core_profiles(self) -> tuple[source.AffectedCoreProfile, ...]:
    return (source.AffectedCoreProfile.PSI,)


class GenericCurrentSourceConfig(source_base.SourceModelBase):
  """Configuration for the GenericCurrentSource.

  Attributes:
    I_generic: total "external" current in A. Used if use_absolute_current=True.
    fraction_of_total_current: total "external" current fraction. Used if
      use_absolute_current=False.
    gaussian_width: width of "external" Gaussian current profile
    gaussian_location: normalized radius of "external" Gaussian current profile
    use_absolute_current: Toggles if external current is provided absolutely or
      as a fraction of Ip.
  """

  model_name: Annotated[Literal['gaussian'], torax_pydantic.JAX_STATIC] = (
      'gaussian'
  )
  I_generic: torax_pydantic.TimeVaryingScalar = torax_pydantic.ValidatedDefault(
      3.0e6
  )
  fraction_of_total_current: torax_pydantic.UnitIntervalTimeVaryingScalar = (
      torax_pydantic.ValidatedDefault(0.2)
  )
  gaussian_width: torax_pydantic.PositiveTimeVaryingScalar = (
      torax_pydantic.ValidatedDefault(0.05)
  )
  gaussian_location: torax_pydantic.UnitIntervalTimeVaryingScalar = (
      torax_pydantic.ValidatedDefault(0.4)
  )
  use_absolute_current: bool = False
  mode: Annotated[runtime_params_lib.Mode, torax_pydantic.JAX_STATIC] = (
      runtime_params_lib.Mode.MODEL_BASED
  )

  @property
  def model_func(self) -> source.SourceProfileFunction:
    return calculate_generic_current

  def build_runtime_params(
      self,
      t: chex.Numeric,
  ) -> RuntimeParams:
    return RuntimeParams(
        prescribed_values=tuple(
            [v.get_value(t) for v in self.prescribed_values]
        ),
        mode=self.mode,
        is_explicit=self.is_explicit,
        I_generic=self.I_generic.get_value(t),
        fraction_of_total_current=self.fraction_of_total_current.get_value(t),
        gaussian_width=self.gaussian_width.get_value(t),
        gaussian_location=self.gaussian_location.get_value(t),
        use_absolute_current=self.use_absolute_current,
    )

  def build_source(self) -> GenericCurrentSource:
    return GenericCurrentSource(model_func=self.model_func)



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for validating the config inputs."""

from collections.abc import Mapping
import functools
import logging
from typing import Annotated, Any, Final, TypeAlias

import numpy as np
import pydantic
from torax._src import constants
from torax._src.torax_pydantic import torax_pydantic

_TOLERANCE: Final[float] = 1e-6


def time_varying_array_defined_at_1(
    time_varying_array: torax_pydantic.TimeVaryingArray,
) -> torax_pydantic.TimeVaryingArray:
  """Validates the input for the TimeVaryingArray."""
  if not time_varying_array.right_boundary_conditions_defined:
    logging.debug("""Not defined at rho=1.0.""")
  return time_varying_array


def time_varying_array_bounded(
    time_varying_array: torax_pydantic.TimeVaryingArray,
    lower_bound: float = -np.inf,
    upper_bound: float = np.inf,
) -> torax_pydantic.TimeVaryingArray:
  """Validates the input for the TimeVaryingArray."""
  for t, (_, values) in time_varying_array.value.items():
    if not np.all(values >= lower_bound):
      raise ValueError(
          f'Some values are smaller than lower bound {lower_bound} at time'
          f' {t}: {values}'
      )
    if not np.all(values <= upper_bound):
      raise ValueError(
          f'Some values are larger than upper bound {upper_bound} at time'
          f' {t}: {values}'
      )
  return time_varying_array


TimeVaryingArrayDefinedAtRightBoundaryAndBounded: TypeAlias = Annotated[
    torax_pydantic.TimeVaryingArray,
    pydantic.AfterValidator(time_varying_array_defined_at_1),
    pydantic.AfterValidator(
        functools.partial(
            time_varying_array_bounded,
            lower_bound=1.0,
        )
    ),
]


def _ion_mixture_before_validator(value: Any) -> Any:
  """Validates the input for the IonMixtureType."""
  if isinstance(value, str):
    return {value: 1.0}
  return value


def _ion_mixture_after_validator(
    value: Mapping[str, torax_pydantic.TimeVaryingScalar],
) -> Mapping[str, torax_pydantic.TimeVaryingScalar]:
  """Validates the input for the IonMixtureType."""
  if not value:
    raise ValueError('The species dictionary cannot be empty.')

  # Check if all species keys are in the allowed list.
  invalid_ion_symbols = set(value.keys()) - constants.ION_SYMBOLS
  if invalid_ion_symbols:
    raise ValueError(
        f'Invalid ion symbols: {invalid_ion_symbols}. Allowed symbols are:'
        f' {constants.ION_SYMBOLS}'
    )

  time_arrays = [v.time for v in value.values()]
  fraction_arrays = [v.value for v in value.values()]

  # Check if all time arrays are equal
  if not all(np.array_equal(time_arrays[0], x) for x in time_arrays[1:]):
    raise ValueError(
        'All time indices for ion mixture fractions must be equal.'
    )

  # Check if the ion fractions sum to 1 at all times
  fraction_sum = np.sum(fraction_arrays, axis=0)
  if not np.allclose(fraction_sum, 1.0, rtol=_TOLERANCE):
    raise ValueError(
        'Fractional concentrations in an IonMixture must sum to 1 at all times.'
    )
  return value


IonMapping: TypeAlias = Annotated[
    Mapping[str, torax_pydantic.TimeVaryingScalar],
    pydantic.BeforeValidator(_ion_mixture_before_validator),
    pydantic.AfterValidator(_ion_mixture_after_validator),
]



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions used for initializing core profiles."""

import dataclasses

import jax
from jax import numpy as jnp
import numpy as np
from torax._src import array_typing
from torax._src import constants
from torax._src import jax_utils
from torax._src import math_utils
from torax._src import state
from torax._src.config import runtime_params_slice
from torax._src.core_profiles import getters
from torax._src.fvm import cell_variable
from torax._src.geometry import geometry
from torax._src.geometry import standard_geometry
from torax._src.neoclassical import neoclassical_models as neoclassical_models_lib
from torax._src.neoclassical.bootstrap_current import base as bootstrap_current_base
from torax._src.physics import psi_calculations
from torax._src.sources import source_models as source_models_lib
from torax._src.sources import source_profile_builders
from torax._src.sources import source_profiles as source_profiles_lib

_trapz = jax.scipy.integrate.trapezoid

# pylint: disable=invalid-name


def initial_core_profiles(
    runtime_params: runtime_params_slice.RuntimeParams,
    geo: geometry.Geometry,
    source_models: source_models_lib.SourceModels,
    neoclassical_models: neoclassical_models_lib.NeoclassicalModels,
) -> state.CoreProfiles:
  """Calculates the initial core profiles.

  Args:
    runtime_params: Runtime parameters at t=t_initial.
    geo: Torus geometry at t=t_initial.
    source_models: All models for TORAX sources/sinks.
    neoclassical_models: All models for neoclassical calculations.

  Returns:
    Initial core profiles.
  """
  T_i = getters.get_updated_ion_temperature(
      runtime_params.profile_conditions, geo
  )
  T_e = getters.get_updated_electron_temperature(
      runtime_params.profile_conditions, geo
  )
  n_e = getters.get_updated_electron_density(
      runtime_params.profile_conditions, geo
  )
  ions = getters.get_updated_ions(runtime_params, geo, n_e, T_e)

  # Set v_loop_lcfs. Two branches:
  # 1. Set the v_loop_lcfs from profile_conditions if using the v_loop BC option
  # 2. Initialize v_loop_lcfs to 0 if using the Ip boundary condition for psi.
  # In case 2, v_loop_lcfs will be updated every timestep based on the psi_lcfs
  # values across the time interval. Since there is is one more time value than
  # time intervals, the v_loop_lcfs time-series is underconstrained. Therefore,
  # we set v_loop_lcfs[0] to v_loop_lcfs[1] when creating the outputs.
  v_loop_lcfs = (
      np.array(runtime_params.profile_conditions.v_loop_lcfs)
      if runtime_params.profile_conditions.use_v_loop_lcfs_boundary_condition
      else np.array(0.0, dtype=jax_utils.get_dtype())
  )

  # Initialise psi and derived quantities to zero before they are calculated.
  psidot = cell_variable.CellVariable(
      value=np.zeros_like(geo.rho),
      dr=geo.drho_norm,
  )
  psi = cell_variable.CellVariable(
      value=np.zeros_like(geo.rho), dr=geo.drho_norm
  )

  core_profiles = state.CoreProfiles(
      T_i=T_i,
      T_e=T_e,
      n_e=n_e,
      n_i=ions.n_i,
      Z_i=ions.Z_i,
      Z_i_face=ions.Z_i_face,
      A_i=ions.A_i,
      n_impurity=ions.n_impurity,
      impurity_fractions=ions.impurity_fractions,
      Z_impurity=ions.Z_impurity,
      Z_impurity_face=ions.Z_impurity_face,
      A_impurity=ions.A_impurity,
      A_impurity_face=ions.A_impurity_face,
      Z_eff=ions.Z_eff,
      Z_eff_face=ions.Z_eff_face,
      psi=psi,
      psidot=psidot,
      q_face=np.zeros_like(geo.rho_face),
      s_face=np.zeros_like(geo.rho_face),
      v_loop_lcfs=v_loop_lcfs,
      sigma=np.zeros_like(geo.rho),
      sigma_face=np.zeros_like(geo.rho_face),
      j_total=np.zeros_like(geo.rho),
      j_total_face=np.zeros_like(geo.rho_face),
      Ip_profile_face=np.zeros_like(geo.rho_face),
  )

  return _init_psi_and_psi_derived(
      runtime_params,
      geo,
      core_profiles,
      source_models,
      neoclassical_models,
  )


def update_psi_from_j(
    Ip: array_typing.FloatScalar,
    geo: geometry.Geometry,
    j_total_hires: jax.Array,
    use_v_loop_lcfs_boundary_condition: bool = False,
) -> cell_variable.CellVariable:
  """Calculates poloidal flux (psi) consistent with plasma current.

  For increased accuracy of psi, a hi-res grid is used, due to the double
    integration. Presently used only for initialization. Therefore Ip is
    a valid source of truth for Ip, even if use_v_loop_lcfs_boundary_condition
    is True.

  Args:
    Ip: Total plasma current [A].
    geo: Torus geometry.
    j_total_hires: High resolution version of j_total [A/m^2].
    use_v_loop_lcfs_boundary_condition: Whether to set the loop voltage from Ip.

  Returns:
    psi: Poloidal flux cell variable.
  """
  y = j_total_hires * geo.spr_hires
  assert y.ndim == 1
  assert geo.rho_hires.ndim == 1
  Ip_profile = math_utils.cumulative_trapezoid(
      y=y, x=geo.rho_hires_norm, initial=0.0
  )
  scale = jnp.concatenate((
      jnp.zeros((1,)),
      (16 * jnp.pi**3 * constants.CONSTANTS.mu0 * geo.Phi_b)
      / (geo.F_hires[1:] * geo.g2g3_over_rhon_hires[1:]),
  ))
  # dpsi_dr on hires cell grid
  dpsi_drhon_hires = scale * Ip_profile

  # psi on hires cell grid
  psi_hires = math_utils.cumulative_trapezoid(
      y=dpsi_drhon_hires, x=geo.rho_hires_norm, initial=0.0
  )

  psi_value = jnp.interp(geo.rho_norm, geo.rho_hires_norm, psi_hires)

  # Set the BCs for psi to ensure the correct Ip
  dpsi_drhonorm_edge = psi_calculations.calculate_psi_grad_constraint_from_Ip(
      Ip,
      geo,
  )

  if use_v_loop_lcfs_boundary_condition:
    # For v_loop_lcfs, we will prescribe a rate of change of psi at the LCFS
    # For the first timestep, we need an initial value for psi at the LCFS, so
    # we set it to match the desired plasma current.
    right_face_grad_constraint = None
    right_face_constraint = (
        psi_value[-1] + dpsi_drhonorm_edge * geo.drho_norm / 2
    )
  else:
    # Use the dpsi/drho calculated above as the right face gradient constraint
    right_face_grad_constraint = dpsi_drhonorm_edge
    right_face_constraint = None

  psi = cell_variable.CellVariable(
      value=psi_value,
      dr=geo.drho_norm,
      right_face_grad_constraint=right_face_grad_constraint,
      right_face_constraint=right_face_constraint,
  )

  return psi


def _init_psi_and_psi_derived(
    runtime_params: runtime_params_slice.RuntimeParams,
    geo: geometry.Geometry,
    core_profiles: state.CoreProfiles,
    source_models: source_models_lib.SourceModels,
    neoclassical_models: neoclassical_models_lib.NeoclassicalModels,
) -> state.CoreProfiles:
  """Initialises psi and currents in core profiles.

  There are three modes of doing this that are supported:
    1. Retrieving psi from the profile conditions.
    2. Retrieving psi from the standard geometry input.
    3. Calculating j according to the nu formula and then
    calculating psi from that. As we are calculating j using a guess for psi,
    this method is iterated to converge to the true psi.

  Args:
    runtime_params: Runtime parameters.
    geo: Torus geometry.
    core_profiles: Core profiles.
    source_models: All TORAX source/sink functions.
    neoclassical_models: All models for neoclassical calculations.

  Returns:
    Refined core profiles.
  """

  # Flag to track if sources have been calculated during psi initialization.
  sources_are_calculated = False

  # Initialize psi source profiles and bootstrap current to all zeros.
  source_profiles = source_profile_builders.build_all_zero_profiles(geo)

  # Case 1: retrieving psi from the profile conditions, using the prescribed
  # profile and Ip
  if runtime_params.profile_conditions.psi is not None:
    # Calculate the dpsi/drho necessary to achieve the given Ip
    dpsi_drhonorm_edge = psi_calculations.calculate_psi_grad_constraint_from_Ip(
        runtime_params.profile_conditions.Ip,
        geo,
    )

    # Set the BCs to ensure the correct Ip
    if runtime_params.profile_conditions.use_v_loop_lcfs_boundary_condition:
      # Extrapolate the value of psi at the LCFS from the dpsi/drho constraint
      # to achieve the desired Ip
      right_face_grad_constraint = None
      right_face_constraint = (
          runtime_params.profile_conditions.psi[-1]
          + dpsi_drhonorm_edge * geo.drho_norm / 2
      )
    else:
      # Use the dpsi/drho calculated above as the right face gradient constraint
      right_face_grad_constraint = dpsi_drhonorm_edge
      right_face_constraint = None

    psi = cell_variable.CellVariable(
        value=runtime_params.profile_conditions.psi,
        right_face_grad_constraint=right_face_grad_constraint,
        right_face_constraint=right_face_constraint,
        dr=geo.drho_norm,
    )

  # Case 2: retrieving psi from the standard geometry input.
  elif (
      isinstance(geo, standard_geometry.StandardGeometry)
      and not runtime_params.profile_conditions.initial_psi_from_j
  ):
    # psi is already provided from a numerical equilibrium, so no need to
    # first calculate currents.
    dpsi_drhonorm_edge = psi_calculations.calculate_psi_grad_constraint_from_Ip(
        runtime_params.profile_conditions.Ip,
        geo,
    )
    # Use the psi from the equilibrium as the right face constraint
    # This has already been made consistent with the desired Ip
    # by make_ip_consistent
    psi = cell_variable.CellVariable(
        value=geo.psi_from_Ip,  # Use psi from equilibrium
        right_face_grad_constraint=None
        if runtime_params.profile_conditions.use_v_loop_lcfs_boundary_condition
        else dpsi_drhonorm_edge,
        right_face_constraint=geo.psi_from_Ip_face[-1]
        if runtime_params.profile_conditions.use_v_loop_lcfs_boundary_condition
        else None,
        dr=geo.drho_norm,
    )

  # Case 3: calculating j according to nu formula and psi from j.
  else:
    # calculate j and psi from the nu formula
    j_total_hires = _get_j_total_hires_with_no_external_sources(
        runtime_params, geo
    )
    psi = update_psi_from_j(
        runtime_params.profile_conditions.Ip,
        geo,
        j_total_hires,
        use_v_loop_lcfs_boundary_condition=runtime_params.profile_conditions.use_v_loop_lcfs_boundary_condition,
    )
    if not (runtime_params.profile_conditions.initial_j_is_total_current):
      # In this branch we require non-inductive currents to determine j_total.
      # The nu formula only provides the Ohmic component of the current.
      # However calculating non-inductive currents requires a non-zero psi.
      # We thus iterate between psi and source calculations, using j_total
      # and psi calculated purely with the nu formula as an initial guess.

      # Initialize iterations
      core_profiles_initial = dataclasses.replace(
          core_profiles,
          psi=psi,
          q_face=psi_calculations.calc_q_face(geo, psi),
          s_face=psi_calculations.calc_s_face(geo, psi),
      )

      # TODO(b/440385263): add tunable iteration number or convergence criteria,
      # and modify python for loop to jax fori loop for the general case.

      # Iterate with non-inductive current source calculations. Stop after 2.
      psi, source_profiles = _iterate_psi_and_sources(
          runtime_params=runtime_params,
          geo=geo,
          core_profiles=core_profiles_initial,
          neoclassical_models=neoclassical_models,
          source_models=source_models,
          source_profiles=source_profiles,
          iterations=2,
      )

      # Mark that sources have been calculated to avoid redundant work.
      sources_are_calculated = True

  # Conclude with completing core_profiles with all psi-dependent profiles.
  core_profiles = _calculate_all_psi_dependent_profiles(
      runtime_params=runtime_params,
      geo=geo,
      psi=psi,
      core_profiles=core_profiles,
      source_profiles=source_profiles,
      source_models=source_models,
      neoclassical_models=neoclassical_models,
      sources_are_calculated=sources_are_calculated,
  )

  return core_profiles


def _calculate_all_psi_dependent_profiles(
    runtime_params: runtime_params_slice.RuntimeParams,
    geo: geometry.Geometry,
    psi: cell_variable.CellVariable,
    core_profiles: state.CoreProfiles,
    source_profiles: source_profiles_lib.SourceProfiles,
    source_models: source_models_lib.SourceModels,
    neoclassical_models: neoclassical_models_lib.NeoclassicalModels,
    sources_are_calculated: bool,
) -> state.CoreProfiles:
  """Supplements core profiles with all other profiles that depend on psi."""
  j_total, j_total_face, Ip_profile_face = psi_calculations.calc_j_total(
      geo, psi
  )

  core_profiles = dataclasses.replace(
      core_profiles,
      psi=psi,
      q_face=psi_calculations.calc_q_face(geo, psi),
      s_face=psi_calculations.calc_s_face(geo, psi),
      j_total=j_total,
      j_total_face=j_total_face,
      Ip_profile_face=Ip_profile_face,
  )
  # Calculate conductivity once we have a consistent set of core profiles
  conductivity = neoclassical_models.conductivity.calculate_conductivity(
      geo,
      core_profiles,
  )

  # Calculate sources if they have not already been calculated.
  if not sources_are_calculated:
    source_profiles = _get_bootstrap_and_standard_source_profiles(
        runtime_params,
        geo,
        core_profiles,
        neoclassical_models,
        source_models,
        source_profiles,
    )

  # psidot calculated here with phibdot=0 in geo, since this is initial
  # conditions and we don't yet have information on geo_t_plus_dt for the
  # phibdot calculation.

  if (
      not runtime_params.numerics.evolve_current
      and runtime_params.profile_conditions.psidot is not None
  ):
    # If psidot is prescribed and psi does not evolve, use prescribed value
    psidot_value = runtime_params.profile_conditions.psidot
  else:
    # Otherwise, calculate psidot from psi sources.
    psi_sources = source_profiles.total_psi_sources(geo)
    psidot_value = psi_calculations.calculate_psidot_from_psi_sources(
        psi_sources=psi_sources,
        sigma=conductivity.sigma,
        resistivity_multiplier=runtime_params.numerics.resistivity_multiplier,
        psi=psi,
        geo=geo,
    )

  # psidot boundary condition. If v_loop_lcfs is not prescribed then we set it
  # to the last calculated psidot for the initialisation since we have no
  # other information.
  v_loop_lcfs = (
      runtime_params.profile_conditions.v_loop_lcfs
      if runtime_params.profile_conditions.use_v_loop_lcfs_boundary_condition
      else psidot_value[-1]
  )
  psidot = dataclasses.replace(
      core_profiles.psidot,
      value=psidot_value,
      right_face_constraint=v_loop_lcfs,
      right_face_grad_constraint=None,
  )
  core_profiles = dataclasses.replace(
      core_profiles,
      psidot=psidot,
      sigma=conductivity.sigma,
      sigma_face=conductivity.sigma_face,
  )
  return core_profiles


def _get_bootstrap_and_standard_source_profiles(
    runtime_params: runtime_params_slice.RuntimeParams,
    geo: geometry.Geometry,
    core_profiles: state.CoreProfiles,
    neoclassical_models: neoclassical_models_lib.NeoclassicalModels,
    source_models: source_models_lib.SourceModels,
    source_profiles: source_profiles_lib.SourceProfiles,
) -> source_profiles_lib.SourceProfiles:
  """Calculates bootstrap current and updates source profiles."""
  source_profile_builders.build_standard_source_profiles(
      runtime_params=runtime_params,
      geo=geo,
      core_profiles=core_profiles,
      source_models=source_models,
      psi_only=True,
      calculate_anyway=True,
      calculated_source_profiles=source_profiles,
  )
  bootstrap_current = (
      neoclassical_models.bootstrap_current.calculate_bootstrap_current(
          runtime_params, geo, core_profiles
      )
  )
  source_profiles = dataclasses.replace(
      source_profiles, bootstrap_current=bootstrap_current
  )
  return source_profiles


def _iterate_psi_and_sources(
    runtime_params: runtime_params_slice.RuntimeParams,
    geo: geometry.Geometry,
    core_profiles: state.CoreProfiles,
    neoclassical_models: neoclassical_models_lib.NeoclassicalModels,
    source_models: source_models_lib.SourceModels,
    source_profiles: source_profiles_lib.SourceProfiles,
    iterations: int,
) -> tuple[cell_variable.CellVariable, source_profiles_lib.SourceProfiles]:
  """Iterates psi and sources to converge to a consistent state."""

  for _ in range(iterations):
    source_profiles = _get_bootstrap_and_standard_source_profiles(
        runtime_params,
        geo,
        core_profiles,
        neoclassical_models,
        source_models,
        source_profiles,
    )
    j_total_hires = get_j_total_hires_with_external_sources(
        runtime_params,
        geo,
        source_profiles.bootstrap_current,
        sum(source_profiles.psi.values()),
    )
    psi = update_psi_from_j(
        runtime_params.profile_conditions.Ip,
        geo,
        j_total_hires,
        use_v_loop_lcfs_boundary_condition=runtime_params.profile_conditions.use_v_loop_lcfs_boundary_condition,
    )
    core_profiles = dataclasses.replace(
        core_profiles,
        psi=psi,
        q_face=psi_calculations.calc_q_face(geo, psi),
        s_face=psi_calculations.calc_s_face(geo, psi),
    )
  return core_profiles.psi, source_profiles


def _get_j_total_hires_with_no_external_sources(
    runtime_params: runtime_params_slice.RuntimeParams,
    geo: geometry.Geometry,
) -> jax.Array:
  """Calculates j_total hires when the total current is given by a formula."""
  Ip = runtime_params.profile_conditions.Ip
  jformula_hires = (
      1 - geo.rho_hires_norm**2
  ) ** runtime_params.profile_conditions.current_profile_nu
  denom = _trapz(jformula_hires * geo.spr_hires, geo.rho_hires_norm)
  Ctot_hires = Ip / denom
  j_total_hires = jformula_hires * Ctot_hires
  return j_total_hires


def get_j_total_hires_with_external_sources(
    runtime_params: runtime_params_slice.RuntimeParams,
    geo: geometry.Geometry,
    bootstrap_current: bootstrap_current_base.BootstrapCurrent,
    external_current: jax.Array,
) -> jax.Array:
  """Calculates j_total hires when the Ohmic current is given by a formula."""
  Ip = runtime_params.profile_conditions.Ip
  psi_current = external_current + bootstrap_current.j_bootstrap

  j_bootstrap_hires = jnp.interp(
      geo.rho_hires, geo.rho_face, bootstrap_current.j_bootstrap_face
  )

  # calculate hi-res "External" current profile (e.g. ECCD) on cell grid.
  external_current_face = math_utils.cell_to_face(
      external_current,
      geo,
      preserved_quantity=math_utils.IntegralPreservationQuantity.SURFACE,
  )
  external_current_hires = jnp.interp(
      geo.rho_hires, geo.rho_face, external_current_face
  )

  # calculate high resolution j_total and Ohmic current profile
  jformula_hires = (
      1 - geo.rho_hires_norm**2
  ) ** runtime_params.profile_conditions.current_profile_nu
  denom = _trapz(jformula_hires * geo.spr_hires, geo.rho_hires_norm)
  I_non_inductive = math_utils.area_integration(psi_current, geo)
  Iohm = Ip - I_non_inductive
  Cohm_hires = Iohm / denom
  j_ohmic_hires = jformula_hires * Cohm_hires
  j_total_hires = j_ohmic_hires + external_current_hires + j_bootstrap_hires
  return j_total_hires



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Profile condition parameters used throughout TORAX simulations."""
import dataclasses
from typing import Annotated, Callable, Final

import chex
import jax
import numpy as np
import pydantic
from torax._src import array_typing
from torax._src.torax_pydantic import torax_pydantic
from typing_extensions import Self

# pylint: disable=invalid-name

# Order of magnitude validations to catch common config errors.
_MIN_IP_AMPS: Final[float] = 1e3
_MIN_DENSITY_M3: Final[float] = 1e10
_MAX_DENSITY_GW: Final[float] = 1e2
_MAX_TEMPERATURE_KEV: Final[float] = 1e3
_MAX_TEMPERATURE_BC_KEV: Final[float] = 5e1


@jax.tree_util.register_dataclass
@dataclasses.dataclass
class RuntimeParams:
  """Prescribed values and boundary conditions for the core profiles."""

  Ip: array_typing.FloatScalar
  v_loop_lcfs: array_typing.FloatScalar
  T_i_right_bc: array_typing.FloatScalar
  T_e_right_bc: array_typing.FloatScalar
  # Temperature profiles defined on the cell grid.
  T_e: array_typing.FloatVector
  T_i: array_typing.FloatVector
  # If provided as array, Psi profile defined on the cell grid.
  psi: array_typing.FloatVector | None
  psidot: array_typing.FloatVector | None
  # Electron density profile on the cell grid.
  n_e: array_typing.FloatVector
  nbar: array_typing.FloatScalar
  n_e_nbar_is_fGW: bool
  n_e_right_bc: array_typing.FloatScalar
  n_e_right_bc_is_fGW: bool
  current_profile_nu: float
  initial_j_is_total_current: bool
  initial_psi_from_j: bool
  normalize_n_e_to_nbar: bool = dataclasses.field(metadata={'static': True})
  use_v_loop_lcfs_boundary_condition: bool = dataclasses.field(
      metadata={'static': True}
  )
  n_e_right_bc_is_absolute: bool = dataclasses.field(metadata={'static': True})


class ProfileConditions(torax_pydantic.BaseModelFrozen):
  """Generic numeric parameters for the simulation.

  The `from_dict(...)` method can accept a dictionary defined by
  https://torax.readthedocs.io/en/latest/configuration.html#profile-conditions.

  Attributes:
    Ip: Total plasma current in A. Note that if Ip_from_parameters=False in
      geometry, then this Ip will be overwritten by values from the geometry
      data. If use_v_loop_lcfs_boundary_condition, only used as an initial
      condition.
    use_v_loop_lcfs_boundary_condition: Boundary condition at LCFS for Vloop ( =
      dspsi_lcfs/dt ). If use_v_loop_lcfs_boundary_condition is True, then the
      specified Vloop at the LCFS is used as the boundary condition for the psi
      equation; otherwise, Ip is used as the boundary condition.
    v_loop_lcfs: Boundary condition at LCFS for Vloop ( = dpsi_lcfs/dt ).
    T_i_right_bc: Temperature boundary conditions at r=a_minor. If this is
      `None` the boundary condition will instead be taken from `T_i` and `T_e`
      at rhon=1.
    T_e_right_bc: Temperature boundary conditions at r=a_minor. If this is
      `None` the boundary condition will instead be taken from `T_i` and `T_e`
      at rhon=1.
    T_i: Prescribed or evolving values for temperature at different times.
    T_e: Prescribed or evolving values for temperature at different times.
    psi: Initial values for psi. If provided, the initial psi will be taken from
      here. Otherwise, the initial psi will be calculated from either the
      geometry or the "current_profile_nu formula" dependant on the
      `initial_psi_from_j` field.
    psidot: Prescribed values for the time derivative of poloidal flux (loop
      voltage). If provided, and if `evolve_current` is False, this prescribed
      `psidot` will be used instead of the internally calculated one. This is
      useful for cases where an unphysical transient `psidot` from the initial
      `psi` condition needs to be overridden.
    n_e: Prescribed or evolving values for electron density at different times.
    normalize_n_e_to_nbar: Whether to renormalize the density profile to have
      the desired line averaged density `nbar`.
    nbar: Line averaged density. In units of m^-3 if n_e_nbar_is_fGW = False. In
      Greenwald fraction if n_e_nbar_is_fGW = True. nGW = Ip/(pi*a^2) with a in
      m, nGW in 10^20 m-3, Ip in MA
    n_e_nbar_is_fGW: Toggle units of nbar
    n_e_right_bc: Density boundary condition for r=a_minor. In units of m^-3 if
      n_e_right_bc_is_fGW = False. In Greenwald fraction if `n_e_right_bc_is_fGW
      = True`. If `n_e_right_bc` is `None` then the boundary condition will
      instead be taken from `n_e` at rho_norm=1. In this case,
      `n_e_right_bc_is_absolute` in the StaticRuntimeParams will be set to
      `False` and n_e_right_bc_is_fGW` will be set to `n_e_nbar_is_fGW`. If
      `n_e_right_bc` is not `None` then `n_e_right_bc_is_absolute` will be set
      to `True`.
    n_e_right_bc_is_fGW: Toggle units of n_e_right_bc.
    current_profile_nu: Peaking factor of "Ohmic" current: j_ohmic = j0*(1 -
      r^2/a^2)^current_profile_nu
    initial_j_is_total_current: Toggles if "Ohmic" current is treated as total
      current upon initialization, or if non-inductive current should be
      included in initial j_total calculation.
    initial_psi_from_j: Toggles if the initial psi calculation is based on the
      "current_profile_nu" current formula, or from the psi available in the
      numerical geometry file. This setting is ignored for the ad-hoc circular
      geometry, which has no numerical geometry.
  """

  Ip: torax_pydantic.TimeVaryingScalar = torax_pydantic.ValidatedDefault(15e6)
  use_v_loop_lcfs_boundary_condition: Annotated[
      bool, torax_pydantic.JAX_STATIC
  ] = False
  v_loop_lcfs: torax_pydantic.TimeVaryingScalar = (
      torax_pydantic.ValidatedDefault(0.0)
  )
  T_i_right_bc: torax_pydantic.PositiveTimeVaryingScalar | None = None
  T_e_right_bc: torax_pydantic.PositiveTimeVaryingScalar | None = None
  T_i: torax_pydantic.PositiveTimeVaryingArray = (
      torax_pydantic.ValidatedDefault({0: {0: 15.0, 1: 1.0}})
  )
  T_e: torax_pydantic.PositiveTimeVaryingArray = (
      torax_pydantic.ValidatedDefault({0: {0: 15.0, 1: 1.0}})
  )
  psi: torax_pydantic.TimeVaryingArray | None = None
  psidot: torax_pydantic.TimeVaryingArray | None = None
  n_e: torax_pydantic.PositiveTimeVaryingArray = (
      torax_pydantic.ValidatedDefault({0: {0: 1.2e20, 1: 0.8e20}})
  )
  normalize_n_e_to_nbar: Annotated[bool, torax_pydantic.JAX_STATIC] = False
  nbar: torax_pydantic.TimeVaryingScalar = torax_pydantic.ValidatedDefault(
      0.85e20
  )
  n_e_nbar_is_fGW: bool = False
  n_e_right_bc: torax_pydantic.TimeVaryingScalar | None = None
  n_e_right_bc_is_fGW: bool = False
  current_profile_nu: float = 1.0
  initial_j_is_total_current: bool = False
  initial_psi_from_j: bool = False

  @pydantic.model_validator(mode='after')
  def after_validator(self) -> Self:
    error_messages = []

    def _sanity_check_profile_boundary_conditions(
        values,
        value_name,
        errors_list,
    ):
      """Check that the profile is defined at rho=1.0 for various cases."""
      if not values.right_boundary_conditions_defined:
        error_message = (
            f'As no right boundary condition was set for {value_name}, the'
            f' profile for {value_name} must include a rho=1.0 boundary'
            ' condition.'
        )
        errors_list.append(error_message)

    if self.T_i_right_bc is None:
      _sanity_check_profile_boundary_conditions(self.T_i, 'T_i', error_messages)
    if self.T_e_right_bc is None:
      _sanity_check_profile_boundary_conditions(self.T_e, 'T_e', error_messages)
    if self.n_e_right_bc is None:
      _sanity_check_profile_boundary_conditions(self.n_e, 'n_e', error_messages)

    # Validate plasma current input order of magnitude.
    if np.any(self.Ip.value < _MIN_IP_AMPS):
      failed_val, time_at_fail = _get_first_failing_value_and_time_or_rho(
          self.Ip.value, self.Ip.time, lambda x: x < _MIN_IP_AMPS
      )
      error_messages.append(
          f'Plasma current Ip at time {time_at_fail}s is {failed_val:.2e} A,'
          f' which is below the minimum threshold of {_MIN_IP_AMPS:.0e} A.'
          ' Possible cause: erroneous input Ip in MA instead of A.'
      )

    # Validate plasma density inputs order of magnitude.
    if not self.n_e_nbar_is_fGW and self.normalize_n_e_to_nbar:
      if np.any(self.nbar.value < _MIN_DENSITY_M3):
        failed_val, time_at_fail = _get_first_failing_value_and_time_or_rho(
            self.nbar.value, self.nbar.time, lambda x: x < _MIN_DENSITY_M3
        )
        error_messages.append(
            'Density inputs set in m^-3 units. Line-averaged density nbar at'
            f' time {time_at_fail}s is {failed_val:.2e} m^-3  which is below'
            f' the minimum threshold of {_MIN_DENSITY_M3:.0e} m^-3. Possible'
            ' cause: erroneous input nbar in normalized units instead of m^-3.'
        )

    if self.n_e_nbar_is_fGW and self.normalize_n_e_to_nbar:
      if np.any(self.nbar.value > _MAX_DENSITY_GW):
        failed_val, time_at_fail = _get_first_failing_value_and_time_or_rho(
            self.nbar.value, self.nbar.time, lambda x: x > _MAX_DENSITY_GW
        )
        error_messages.append(
            'Density inputs set in Greenwald fraction. Line-averaged density'
            f' nbar time {time_at_fail}s is {failed_val:.2e} fGW which is'
            f' above the maximum threshold of {_MAX_DENSITY_GW:.0e} fGW.'
            ' Possible cause: erroneous input nbar in m^-3 instead of fGW.'
        )

    if not self.n_e_nbar_is_fGW and not self.normalize_n_e_to_nbar:
      for n_e_time, n_e_value in self.n_e.value.items():
        if np.any(n_e_value[1] < _MIN_DENSITY_M3):
          failed_val, rho_norm_at_fail = (
              _get_first_failing_value_and_time_or_rho(
                  n_e_value[1], n_e_value[0], lambda x: x < _MIN_DENSITY_M3
              )
          )
          error_messages.append(
              f'Density inputs set in m^-3 units. n_e at time {n_e_time}s and'
              f' rho_norm {rho_norm_at_fail} is {failed_val:.2e} m^-3  which is'
              f' below the minimum threshold of {_MIN_DENSITY_M3:.0e} m^-3.'
              ' Possible cause: erroneous input n_e in normalized units instead'
              ' of m^-3.'
          )
        break

    if self.n_e_nbar_is_fGW and not self.normalize_n_e_to_nbar:
      for n_e_time, n_e_value in self.n_e.value.items():
        if np.any(n_e_value[1] > _MAX_DENSITY_GW):
          (
              failed_val,
              rho_norm_at_fail,
          ) = _get_first_failing_value_and_time_or_rho(
              n_e_value[1], n_e_value[0], lambda x: x > _MAX_DENSITY_GW
          )
          error_messages.append(
              'Density inputs set in Greenwald fraction. n_e at time'
              f' {n_e_time}s and rho_norm {rho_norm_at_fail} is'
              f' {failed_val:.2e} m^-3  which is above the maximum threshold of'
              f' {_MAX_DENSITY_GW:.0e} fGW. Possible cause: erroneous input n_e'
              ' in m^-3 instead of fGW.'
          )
          break

    if self.n_e_right_bc is not None:
      if not self.n_e_right_bc_is_fGW:
        if np.any(self.n_e_right_bc.value < _MIN_DENSITY_M3):
          failed_val, time_at_fail = _get_first_failing_value_and_time_or_rho(
              self.n_e_right_bc.value,
              self.n_e_right_bc.time,
              lambda x: x < _MIN_DENSITY_M3,
          )
          error_messages.append(
              'Density boundary condition inputs set in m^-3 units,'
              f' n_e_right_bc at time {time_at_fail}s is {failed_val:.2e} m^-3'
              f' which is below the minimum threshold of {_MIN_DENSITY_M3:.0e}'
              ' m^-3. Possible cause: erroneous input n_e_right_bc in'
              ' normalized units instead of m^-3.'
          )
      else:
        if np.any(self.n_e_right_bc.value > _MAX_DENSITY_GW):
          failed_val, time_at_fail = _get_first_failing_value_and_time_or_rho(
              self.n_e_right_bc.value,
              self.n_e_right_bc.time,
              lambda x: x > _MAX_DENSITY_GW,
          )
          error_messages.append(
              'Density boundary condition inputs set in Greenwald fraction,'
              f' n_e_right_bc at time {time_at_fail}s is {failed_val:.2e} fGW'
              f' which is above the maximum threshold of {_MAX_DENSITY_GW:.0e}'
              ' fGW. Possible cause: erroneous input n_e_right_bc in m^-3'
              ' instead of fGW.'
          )

    # Validate temperature inputs order of magnitude.
    for T_e_time, T_e_value in self.T_e.value.items():
      if np.any(T_e_value[1] > _MAX_TEMPERATURE_KEV):
        (
            failed_val,
            rho_norm_at_fail,
        ) = _get_first_failing_value_and_time_or_rho(
            T_e_value[1], T_e_value[0], lambda x: x > _MAX_TEMPERATURE_KEV
        )
        error_messages.append(
            f'T_e at time {T_e_time}s and rho_norm {rho_norm_at_fail} is'
            f' {failed_val:.2e} which is above the maximum threshold of'
            f' {_MAX_TEMPERATURE_KEV:.0e} keV. Possible cause: erroneous'
            ' input T_e in eV instead of keV.'
        )
        break

    for T_i_time, T_i_value in self.T_i.value.items():
      if np.any(T_i_value[1] > _MAX_TEMPERATURE_KEV):
        (
            failed_val,
            rho_norm_at_fail,
        ) = _get_first_failing_value_and_time_or_rho(
            T_i_value[1], T_i_value[0], lambda x: x > _MAX_TEMPERATURE_KEV
        )
        error_messages.append(
            f'T_i at time {T_i_time}s and rho_norm {rho_norm_at_fail} is'
            f' {failed_val:.2e} which is above the maximum threshold of'
            f' {_MAX_TEMPERATURE_KEV:.0e} keV. Possible cause: erroneous'
            ' input T_i in eV instead of keV.'
        )
        break

    if self.T_e_right_bc is not None:
      if np.any(self.T_e_right_bc.value > _MAX_TEMPERATURE_BC_KEV):
        failed_val, time_at_fail = _get_first_failing_value_and_time_or_rho(
            self.T_e_right_bc.value,
            self.T_e_right_bc.time,
            lambda x: x > _MAX_TEMPERATURE_BC_KEV,
        )
        error_messages.append(
            f'T_e_right_bc at time {time_at_fail}s is {failed_val:.2e} keV'
            ' which is above the maximum threshold of'
            f' {_MAX_TEMPERATURE_BC_KEV:.0e} keV. Possible cause: erroneous'
            ' input T_e_right_bc in eV instead of keV.'
        )

    if self.T_i_right_bc is not None:
      if np.any(self.T_i_right_bc.value > _MAX_TEMPERATURE_BC_KEV):
        failed_val, time_at_fail = _get_first_failing_value_and_time_or_rho(
            self.T_i_right_bc.value,
            self.T_i_right_bc.time,
            lambda x: x > _MAX_TEMPERATURE_BC_KEV,
        )
        error_messages.append(
            f'T_i_right_bc at time {time_at_fail}s is {failed_val:.2e} keV'
            ' which is above the maximum threshold of'
            f' {_MAX_TEMPERATURE_BC_KEV:.0e} keV. Possible cause: erroneous'
            ' input T_i_right_bc in eV instead of keV.'
        )

    if error_messages:
      error_message_preamble = (
          f'{len(error_messages)} errors were found in profile conditions'
          ' config:\n\n'
      )
      final_error_message = '\n\n'.join(error_messages)
      raise ValueError(error_message_preamble + final_error_message)

    return self

  def build_runtime_params(self, t: chex.Numeric) -> RuntimeParams:
    """Builds a RuntimeParams object for time t."""

    runtime_params = {
        x.name: getattr(self, x.name)
        for x in dataclasses.fields(RuntimeParams)
        if x.name != 'n_e_right_bc_is_absolute'
    }

    if self.T_e_right_bc is None:
      runtime_params['T_e_right_bc'] = self.T_e.get_value(
          t, grid_type='face_right'
      )

    if self.T_i_right_bc is None:
      runtime_params['T_i_right_bc'] = self.T_i.get_value(
          t, grid_type='face_right'
      )

    if self.n_e_right_bc is None:
      runtime_params['n_e_right_bc'] = self.n_e.get_value(
          t, grid_type='face_right'
      )
      runtime_params['n_e_right_bc_is_fGW'] = self.n_e_nbar_is_fGW
      runtime_params['n_e_right_bc_is_absolute'] = False
    else:
      runtime_params['n_e_right_bc_is_absolute'] = True

    def _get_value(x):
      if isinstance(
          x, (torax_pydantic.TimeVaryingScalar, torax_pydantic.TimeVaryingArray)
      ):
        return x.get_value(t)
      else:
        return x

    runtime_params = {k: _get_value(v) for k, v in runtime_params.items()}
    return RuntimeParams(**runtime_params)


def _get_first_failing_value_and_time_or_rho(
    values: np.ndarray,
    time_or_rho: np.ndarray,
    comparator: Callable[[np.ndarray], np.ndarray],
) -> tuple[float, float]:
  """Returns the first failing value and time or rho for a given comparator.

  If values came from a TimeVaryingScalar, then time_or_rho will be the time.
  If values came from a TimeVaryingArray, then time_or_rho will be the rho.

  Args:
    values: The values to check.
    time_or_rho: The time or rho associated with the values.
    comparator: A function that takes an array of values and returns an array of
      bools indicating whether each value passes the check.

  Returns:
    A tuple of (first_failing_value, time_or_rho)
  """
  fail_indices = np.where(comparator(values))[0]
  if fail_indices.size == 0:
    raise ValueError(
        '_get_first_failing_time was called with no failing values.'
    )
  return values[fail_indices[0]], time_or_rho[fail_indices[0]]



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The PedestalModel abstract base class.

The pedestal model calculates quantities relevant to the pedestal.
"""
import abc
import dataclasses

import jax
import jax.numpy as jnp
from torax._src import array_typing
from torax._src import state
from torax._src.config import runtime_params_slice
from torax._src.geometry import geometry

# pylint: disable=invalid-name
# Using physics notation naming convention


@jax.tree_util.register_dataclass
@dataclasses.dataclass(frozen=True)
class PedestalModelOutput:
  """Output of the PedestalModel."""

  # The location of the pedestal.
  rho_norm_ped_top: array_typing.FloatScalar
  # The index of the pedestal in rho_norm.
  rho_norm_ped_top_idx: array_typing.IntScalar
  # The ion temperature at the pedestal.
  T_i_ped: array_typing.FloatScalar
  # The electron temperature at the pedestal.
  T_e_ped: array_typing.FloatScalar
  # The electron density at the pedestal in units 10^-3.
  n_e_ped: array_typing.FloatScalar


class PedestalModel(abc.ABC):
  """Calculates temperature and density of the pedestal.

  Subclass responsbilities:
  - Must set _frozen = True at the end of the subclass __init__ method to
    activate immutability.
  """

  def __setattr__(self, attr, value):
    # pylint: disable=g-doc-args
    # pylint: disable=g-doc-return-or-yield
    """Override __setattr__ to make the class (sort of) immutable.

    Note that you can still do obj.field.subfield = x, so it is not true
    immutability, but this to helps to avoid some careless errors.
    """
    if getattr(self, "_frozen", False):
      raise AttributeError("PedestalModels are immutable.")
    return super().__setattr__(attr, value)

  def __call__(
      self,
      runtime_params: runtime_params_slice.RuntimeParams,
      geo: geometry.Geometry,
      core_profiles: state.CoreProfiles,
  ) -> PedestalModelOutput:
    if not getattr(self, "_frozen", False):
      raise RuntimeError(
          f"Subclass implementation {type(self)} forgot to "
          "freeze at the end of __init__."
      )

    return jax.lax.cond(
        runtime_params.pedestal.set_pedestal,
        lambda: self._call_implementation(runtime_params, geo, core_profiles),
        # Set the pedestal location to infinite to indicate that the pedestal is
        # not present.
        # Set the index to outside of bounds of the mesh to indicate that the
        # pedestal is not present.
        lambda: PedestalModelOutput(
            rho_norm_ped_top=jnp.inf,
            T_i_ped=0.0,
            T_e_ped=0.0,
            n_e_ped=0.0,
            rho_norm_ped_top_idx=geo.torax_mesh.nx,
        ),
    )

  @abc.abstractmethod
  def _call_implementation(
      self,
      runtime_params: runtime_params_slice.RuntimeParams,
      geo: geometry.Geometry,
      core_profiles: state.CoreProfiles,
  ) -> PedestalModelOutput:
    """Calculate the pedestal values."""

  @abc.abstractmethod
  def __hash__(self) -> int:
    """Hash function for the pedestal model.

    Needed for jax.jit caching to work.
    """
    ...

  @abc.abstractmethod
  def __eq__(self, other) -> bool:
    """Equality function for the pedestal model."""
    ...



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A solver that implements the sawtooth trigger and redistribution."""

import dataclasses

import jax
from torax._src import state
from torax._src.config import runtime_params_slice
from torax._src.core_profiles import convertors
from torax._src.fvm import cell_variable
from torax._src.geometry import geometry
from torax._src.solver import solver
from torax._src.sources import source_profiles as source_profiles_lib


# TODO(b/414537757). Sawtooth extensions.
# a. Full and incomplete Kadomtsev redistribution model.
# b. Porcelli model with free parameters and fast ion sensitivities.
# c. "Smooth" version that can work with forward-sensitivity-analysis and
#    stationary-state applications without the need for averaging.
class SawtoothSolver(solver.Solver):
  """Sawtooth trigger and redistribution, and carries out sawtooth step."""

  def _x_new(
      self,
      dt: jax.Array,
      runtime_params_t: runtime_params_slice.RuntimeParams,
      runtime_params_t_plus_dt: runtime_params_slice.RuntimeParams,
      geo_t: geometry.Geometry,
      geo_t_plus_dt: geometry.Geometry,
      core_profiles_t: state.CoreProfiles,
      core_profiles_t_plus_dt: state.CoreProfiles,
      explicit_source_profiles: source_profiles_lib.SourceProfiles,
      evolving_names: tuple[str, ...],
  ) -> tuple[
      tuple[cell_variable.CellVariable, ...],
      state.SolverNumericOutputs,
  ]:
    """Applies the sawtooth model and outputs new state attributes if triggered.

    If the trigger model indicates a crash has been triggered, an
    instantaneous redistribution model is applied. New state attributes
    following a short (configurable) dt are returned. Beyond the sawtooth
    redistribution, core_profiles are further updated by: the psidot assumed at
    time t; the new boundary conditions; sources, transport, geo, and pedestal
    outputs consistent with the new core_profiles at time t_plus_crash_dt.

    Args:
      dt: Sawtooth step duration.
      runtime_params_t: Runtime parameters at time t.
      runtime_params_t_plus_dt: Runtime parameters at time t + crash_dt.
      geo_t: Geometry at time t.
      geo_t_plus_dt: Geometry at time t + crash_dt.
      core_profiles_t: Core profiles at time t.
      core_profiles_t_plus_dt: Core profiles containing boundary conditions and
        prescribed profiles at time t + crash_dt.
      explicit_source_profiles: Explicit source profiles at time t.
      evolving_names: Names of evolving variables.

    Returns:
      Updated tuple of evolving CellVariables from CoreProfiles
      SolverNumericOutputs indicating a sawtooth crash.
    """
    sawtooth_models = self.physics_models.mhd_models.sawtooth_models
    if sawtooth_models is None:
      raise ValueError('Sawtooth model is None.')

    trigger_sawtooth, rho_norm_q1 = sawtooth_models.trigger_model(
        runtime_params_t,
        geo_t,
        core_profiles_t,
    )

    def _redistribute_state() -> tuple[
        tuple[cell_variable.CellVariable, ...],
        state.SolverNumericOutputs,
    ]:

      redistributed_core_profiles = sawtooth_models.redistribution_model(
          rho_norm_q1,
          runtime_params_t,
          geo_t,
          core_profiles_t,
      )

      # Evolve the psi profile over the sawtooth time.
      # Redistribution maintains the same psi boundary condition. However,
      # over the course of the sawtooth time, the central solenoid must still
      # modify the psi profile. Since we don't calculate the psi PDE here, we
      # assume that for the short sawtooth time we can use the psidot from the
      # beginning of the step interval. This updates the bulk values. Later, the
      # boundary conditions are also updated at time t_plus_dt when
      # using `updaters.update_all_core_profiles_after_step`.
      evolved_psi_redistributed_value = (
          redistributed_core_profiles.psi.value
          + core_profiles_t.psidot.value * dt
      )
      evolved_core_profiles = dataclasses.replace(
          redistributed_core_profiles,
          psi=dataclasses.replace(
              redistributed_core_profiles.psi,
              value=evolved_psi_redistributed_value,
          ),
      )

      x_post_step = convertors.core_profiles_to_solver_x_tuple(
          evolved_core_profiles, evolving_names
      )

      solver_numeric_outputs_post_step = state.SolverNumericOutputs(
          sawtooth_crash=True
      )

      return (
          x_post_step,
          solver_numeric_outputs_post_step,
      )

    # Return redistributed state attributes if triggered, otherwise return
    # unchanged state attributes.
    return jax.lax.cond(
        trigger_sawtooth,
        _redistribute_state,
        lambda: (
            tuple([getattr(core_profiles_t, name) for name in evolving_names]),
            state.SolverNumericOutputs(),
        ),
    )
